# generated by datamodel-codegen:
#   filename:  artifacts.json
#   timestamp: 2025-03-03T13:22:20+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from truefoundry.pydantic_v1 import BaseModel, Field, conint, constr


class Method(str, Enum):
    """
    HTTP Method for the tool call HTTP request
    """

    get = "get"
    post = "post"
    put = "put"
    delete = "delete"
    patch = "patch"


class ArtifactPath(BaseModel):
    src: str = Field(..., description="Local file or folder path")
    dest: Optional[str] = Field(
        None,
        description="Relative path where the file or folder will be uploaded to in the artifact",
    )


class BaseArtifactVersion(BaseModel):
    name: Optional[constr(regex=r"^[A-Za-z0-9_\-]+$", max_length=256)] = Field(
        None, description="Name of the entity"
    )
    description: Optional[constr(max_length=512)] = Field(None, description="")
    metadata: Dict[str, Any] = Field(
        ...,
        description='Key value metadata. Should be valid JSON. For e.g. `{"business-unit": "sales", "quality": "good", "rating": 4.5}`',
    )
    version_alias: Optional[
        constr(regex=r"^v[a-zA-Z0-9.-]*([a-zA-Z0-9]+)$", max_length=128)
    ] = Field(
        None,
        description="Version alias is alternate, ideally human readable, version string to reference an artifact version. It should start with `v` followed by alphanumeric and it can include `.` and `-` in between (e.g. `v1.0.0`, `v1-prod`, `v3-dev`, etc)",
    )
    ml_repo: Optional[constr(regex=r"^[a-zA-Z][a-zA-Z0-9\-]{1,98}[a-zA-Z0-9]$")] = (
        Field(None, description="Name of the ML Repo")
    )
    version: Optional[conint(ge=1)] = Field(None, description="Version of the entity")


class MimeType(str, Enum):
    """
    MIME type of the content
    """

    text_plain = "text/plain"
    application_json = "application/json"
    image_png = "image/png"
    image_jpeg = "image/jpeg"
    application_x_directory = "application/x-directory"


class BlobStorageReference(BaseModel):
    """
    A reference to content offloaded to blob storage
    """

    type: Literal["blob-storage"] = Field(..., description="")
    path: constr(regex=r"^.{1,}$") = Field(
        ..., description="Path to the content in blob storage"
    )
    mime_type: MimeType = Field(..., description="MIME type of the content")


class Collaborator(BaseModel):
    subject: str = Field(
        ...,
        description="Fully Qualified Name of the subject. eg: user:email or team:teamname",
    )
    role_id: str = Field(..., description="Role ID for the resource")


class ExternalBlobStorageSource(BaseModel):
    type: Literal["external"] = Field(..., description="Type of the source")
    uri: constr(regex=r"^(s3|gs|wasbs)://.*") = Field(
        ...,
        description="URI referencing a path in the blob storage bucket linked to the MLRepo",
    )


class FastAIFramework(BaseModel):
    type: Literal["fastai"] = Field(..., description="Type of the framework")


class GluonFramework(BaseModel):
    type: Literal["gluon"] = Field(..., description="Type of the framework")


class H2OFramework(BaseModel):
    type: Literal["h2o"] = Field(..., description="Type of the framework")


class ImageUrl(BaseModel):
    """
    Image URL linking to the image
    """

    url: Union[
        constr(
            regex=r"^\b((https?://)?(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}(?:[-a-zA-Z0-9()@:%_\+.~#?&/=]*)|data:image/[a-zA-Z]+;base64,[a-zA-Z0-9+/=\s]+)$"
        ),
        BlobStorageReference,
    ] = Field(..., description="Image URL linking to the image")


class ImageContentPart(BaseModel):
    """
    Image content for the message
    """

    type: Literal["image_url"] = Field(..., description="Type of the content part")
    image_url: ImageUrl = Field(..., description="Image URL linking to the image")


class KerasFramework(BaseModel):
    type: Literal["keras"] = Field(..., description="Type of the framework")


class LightGBMFramework(BaseModel):
    type: Literal["lightgbm"] = Field(..., description="Type of the framework")


class LocalArtifactSource(BaseModel):
    type: Literal["local"] = Field(..., description="Type of the source")
    paths: List[ArtifactPath] = Field(
        ...,
        description="Array of ArtifactPath objects representing the source and destination paths",
    )


class LocalModelSource(BaseModel):
    type: Literal["local"] = Field(..., description="Type of the source")
    file_or_folder: str = Field(..., description="Path to the model file or folder")


class MLRepo(BaseModel):
    """
    MLRepo is a repository ML training runs that log params, metrics, plots, images and versioned entities like artifacts, models, prompts, tools, agents
    """

    type: Literal["ml-repo"] = Field(..., description="")
    name: constr(regex=r"^[a-zA-Z][a-zA-Z0-9\-]{1,98}[a-zA-Z0-9]$") = Field(
        ..., description="Name of the ML Repo."
    )
    description: Optional[constr(regex=r"^.{0,500}$")] = Field(
        None, description="Description for the ML Repo."
    )
    storage_integration_fqn: str = Field(
        ...,
        description="Storage Integration to store artifacts and models. A storage integration represents a connected blob storage like AWS S3 / Azure Blob Storage / Google Cloud Storage.",
    )
    collaborators: List[Collaborator] = Field(
        ..., description="Users and Teams that have access to MLRepo"
    )


class Parameters(BaseModel):
    """
    Parameters to pass to the model when generating
    """

    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_k: Optional[float] = None
    top_p: Optional[float] = None
    stop: Optional[Union[List[str], str]] = None


class ModelConfiguration(BaseModel):
    """
    Add a model from Gateway along with parameters to be used for chat completion
    """

    provider: str = Field(
        ...,
        description="Name of the provider, must be one of the integration providers configured for the Gateway",
    )
    model: constr(regex=r"^[a-zA-Z][a-zA-Z0-9\-]{1,30}[a-zA-Z0-9]$") = Field(
        ...,
        description="Name of the model to be used for generations. This model should be available in the provider",
    )
    parameters: Optional[Parameters] = Field(
        None, description="Parameters to pass to the model when generating"
    )
    extra_parameters: Optional[Dict[str, Any]] = Field(
        None,
        description="Arbitrary extra parameters to pass to the model when generating",
    )


class ModelVersionEnvironment(BaseModel):
    """
    Runtime environment describing python version and dependencies for running the model
    """

    python_version: Optional[constr(regex=r"^\d+(\.\d+){1,2}([\-\.a-z0-9]+)?$")] = (
        Field(None, description="Python version to use when running the model")
    )
    pip_packages: Optional[List[str]] = Field(
        None, description="pip dependencies needed for running the model"
    )


class ONNXFramework(BaseModel):
    type: Literal["onnx"] = Field(..., description="Type of the framework")


class PaddleFramework(BaseModel):
    type: Literal["paddle"] = Field(..., description="Type of the framework")


class PyTorchFramework(BaseModel):
    type: Literal["pytorch"] = Field(..., description="Type of the framework")


class InferMethodName(str, Enum):
    """
    Name of the method used for inference
    """

    predict = "predict"
    predict_proba = "predict_proba"


class SklearnModelSchema(BaseModel):
    infer_method_name: InferMethodName = Field(
        ..., description="Name of the method used for inference"
    )
    inputs: List[Dict[str, Any]] = Field(..., description="Schema of the input")
    outputs: List[Dict[str, Any]] = Field(..., description="Schema of the output")


class SklearnSerializationFormat(str, Enum):
    """
    Serialization format used to save the sklearn model
    """

    cloudpickle = "cloudpickle"
    joblib = "joblib"
    pickle = "pickle"


class SpaCyFramework(BaseModel):
    type: Literal["spacy"] = Field(..., description="Type of the framework")


class StatsModelsFramework(BaseModel):
    type: Literal["statsmodels"] = Field(..., description="Type of the framework")


class SystemMessage(BaseModel):
    """
    System message for the chat
    """

    role: Literal["system"] = Field(..., description="Role of the message")
    content: Union[constr(regex=r"^.[\s\S]*$"), BlobStorageReference] = Field(
        ..., description="Text content for the system message"
    )
    name: Optional[str] = Field(None, description="Name of the system")


class TensorFlowFramework(BaseModel):
    type: Literal["tensorflow"] = Field(..., description="Type of the framework")


class TextContentPart(BaseModel):
    """
    Text content for the message
    """

    type: Literal["text"] = Field(..., description="Type of the content part")
    text: Union[constr(regex=r"^.[\s\S]*$"), BlobStorageReference] = Field(
        ..., description="Text content for the message"
    )


class LibraryName(str, Enum):
    """
    Name of the library for the framework
    """

    transformers = "transformers"
    sentence_transformers = "sentence-transformers"
    diffusers = "diffusers"


class TransformersFramework(BaseModel):
    type: Literal["transformers"] = Field(..., description="Type of the framework")
    library_name: Optional[LibraryName] = Field(
        "transformers", description="Name of the library for the framework"
    )
    pipeline_tag: Optional[str] = Field(
        None,
        description="The `pipeline()` task this model can be used with e.g. `text-generation`. See [huggingface docs](https://huggingface.co/docs/transformers/main/en/main_classes/pipelines#transformers.pipeline.task) for all possible values",
    )
    base_model: Optional[str] = Field(
        None,
        description="Base model Id from HuggingFace Hub. If this is a finetuned model, this points to the base model id used for finetuning.",
    )


class TrueFoundryManagedSource(BaseModel):
    type: Literal["truefoundry"] = Field(..., description="Type of the source")
    uri: Optional[constr(regex=r"^(s3|gs|wasbs)://.*")] = Field(None, description="")


class UserMessage(BaseModel):
    """
    User message turn
    """

    role: Literal["user"] = Field(..., description="Role of the message")
    content: Union[
        constr(regex=r"^.[\s\S]*$"),
        BlobStorageReference,
        List[Union[TextContentPart, ImageContentPart]],
    ] = Field(
        ..., description="Content of the user message. can be a mix of text and images"
    )
    name: Optional[str] = Field(
        None, description="Name of the user this message is from"
    )


class XGBoostModelSchema(BaseModel):
    """
    Schema of the XGBoost model
    """

    infer_method_name: Literal["predict"] = Field(
        ..., description="Name of the method used for inference"
    )
    inputs: List[Dict[str, Any]] = Field(..., description="Schema of the input")
    outputs: List[Dict[str, Any]] = Field(..., description="Schema of the output")


class XGBoostSerializationFormat(str, Enum):
    """
    Serialization format used to save the xgboost model
    """

    cloudpickle = "cloudpickle"
    joblib = "joblib"
    pickle = "pickle"
    json = "json"


class Agent(BaseArtifactVersion):
    type: Literal["agent"] = Field(..., description="Type of the entity")
    available_tools: List[constr(min_length=1)] = Field(
        ..., description="Tools available to the agent"
    )
    goal: constr(min_length=1, max_length=128) = Field(
        ...,
        description="Short form description. Will be used as `description` when this agent is used as a tool.",
    )
    instruction: constr(min_length=1, max_length=2620) = Field(
        ..., description="Instructions for the agent to follow to achieve the goal"
    )
    model_id: constr(min_length=1) = Field(
        ..., description="Model to use when running the agent"
    )


class AgentOpenAPITool(BaseArtifactVersion):
    type: Literal["openapi-tool"] = Field(..., description="Type of the agent")
    openapi_spec: Union[BlobStorageReference, Dict[str, Any]] = Field(
        ...,
        description="OpenAPI Spec for the tool describing the API, endpoints and parameters. [Sample OpenAPI Spec Link](https://assets.production.truefoundry.com/sample-openapi.json)",
    )
    base_url: constr(
        regex=r"^(https?://)?(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}(?:[-a-zA-Z0-9()@:%_\+.~#?&/=]*)$"
    ) = Field(
        ...,
        description="HTTP endpoint where the API is hosted for the tools. E.g. `https://api.example.com/v1`",
    )
    path: constr(min_length=1) = Field(
        ...,
        description="API Route Path for the tool call HTTP request. E.g. `GET /weather`",
    )
    method: Method = Field(
        ..., description="HTTP Method for the tool call HTTP request"
    )
    headers: Optional[Dict[str, str]] = Field(
        None,
        description="HTTP Headers for the tool call HTTP request. E.g. `Authorization: Bearer <token>`",
    )


class AgentOpenAPIToolWithFQN(AgentOpenAPITool):
    id: str = Field(..., description="ID of the agent tool")
    fqn: str = Field(..., description="FQN of the agent tool")


class AgentWithFQN(Agent):
    id: str = Field(..., description="ID of the agent")
    fqn: str = Field(..., description="FQN of the agent")


class ArtifactVersion(BaseArtifactVersion):
    """
    Log a new Artifact Version containing files and folders with metadata
    """

    type: Literal["artifact-version"] = Field(..., description="Artifact Version")
    source: Union[
        TrueFoundryManagedSource, ExternalBlobStorageSource, LocalArtifactSource
    ] = Field(..., description="")
    step: conint(ge=0) = Field(
        0,
        description="Step/Epoch number in an iterative training loop the artifact version was created. Generally useful when logging a model version from a MLRepo Run",
    )
    run_id: Optional[str] = Field(
        None, description="ID of the MLRepo Run that generated the artifact version"
    )


class AssistantMessage(BaseModel):
    """
    Assistant message turn
    """

    role: Literal["assistant"] = Field(..., description="Role of the message")
    content: Union[constr(regex=r"^.[\s\S]*$"), BlobStorageReference] = Field(
        ..., description="Text content for the assistant message"
    )
    name: Optional[str] = Field(
        None, description="Name of the assistant this message is from"
    )


class ChatPrompt(BaseArtifactVersion):
    """
    Fill in the information for opening a new chat prompt
    """

    type: Literal["chat_prompt"] = Field(..., description="Type of the prompt")
    messages: List[Union[SystemMessage, AssistantMessage, UserMessage]] = Field(
        ..., description="List of messages in the chat conversation, must be non-empty"
    )
    variables: Optional[Dict[str, Optional[str]]] = Field(
        None,
        description="Variables referenced in messages and that can be replaced when running generation",
    )
    model_configuration: ModelConfiguration


class DataDirectory(BaseModel):
    type: Literal["data-dir"] = Field(..., description="")
    name: constr(regex=r"^[A-Za-z0-9_\-]+$", max_length=256) = Field(
        ..., description="Name of the data directory"
    )
    ml_repo: constr(regex=r"^[a-zA-Z][a-zA-Z0-9\-]{1,98}[a-zA-Z0-9]$") = Field(
        ..., description="Name of the ML Repo to create the data directory under"
    )
    description: Optional[constr(max_length=512)] = Field(
        None, description="Description for the data directory"
    )
    metadata: Dict[str, Any] = Field(
        ...,
        description='Key Value metadata. Should be valid JSON. For e.g. `{"business-unit": "sales", "quality": "good", "rating": 4.5}`',
    )
    source: Union[TrueFoundryManagedSource, ExternalBlobStorageSource] = Field(
        ..., description=""
    )


class SklearnFramework(BaseModel):
    type: Literal["sklearn"] = Field(..., description="Type of the framework")
    model_filepath: Optional[str] = Field(
        None,
        description="Relative path to the model file in the model version contents",
    )
    serialization_format: Optional[SklearnSerializationFormat] = None
    model_schema: Optional[SklearnModelSchema] = None


class XGBoostFramework(BaseModel):
    type: Literal["xgboost"] = Field(..., description="Type of the framework")
    serialization_format: Optional[XGBoostSerializationFormat] = None
    model_filepath: Optional[str] = Field(
        None,
        description="Relative path to the model file in the model version contents",
    )
    model_schema: Optional[XGBoostModelSchema] = None


class AgentApp(BaseModel):
    type: Literal["agent-app"] = Field(..., description="Type of the app")
    tools: List[AgentOpenAPIToolWithFQN] = Field(
        ..., description="Tools available to the Agent app"
    )
    agents: List[AgentWithFQN] = Field(
        ..., description="Agents available to the Agent app"
    )
    root_agent: constr(min_length=1) = Field(
        ..., description="Root Agent for the app. This will be the first agent invoked"
    )


class ModelVersion(BaseArtifactVersion):
    """
    Log a new Model Version containing model files and folders with metadata
    """

    type: Literal["model-version"] = Field(..., description="Model Version")
    source: Union[
        TrueFoundryManagedSource, ExternalBlobStorageSource, LocalModelSource
    ] = Field(..., description="")
    framework: Optional[
        Union[
            TransformersFramework,
            TensorFlowFramework,
            SklearnFramework,
            PyTorchFramework,
            KerasFramework,
            XGBoostFramework,
            LightGBMFramework,
            FastAIFramework,
            H2OFramework,
            ONNXFramework,
            SpaCyFramework,
            StatsModelsFramework,
            GluonFramework,
            PaddleFramework,
        ]
    ] = Field(
        None,
        description="Framework for the model version like Transformers, PyTorch, Sklearn, Xgboost etc with framework specific metadata. This will be used to infer model deployment configuration",
    )
    environment: Optional[ModelVersionEnvironment] = None
    step: conint(ge=0) = Field(
        0,
        description="Step/Epoch number in an iterative training loop the model version was created. Generally useful when logging a model version from a MLRepo Run",
    )
    run_id: Optional[str] = Field(
        None, description="ID of the MLRepo Run that generated the model version"
    )


class VersionedArtifactType(BaseModel):
    __root__: Union[
        ArtifactVersion, ModelVersion, ChatPrompt, AgentOpenAPITool, Agent, AgentApp
    ]



import ast
import io
import json
import re
from typing import Any, Callable, Dict, List, Optional

from rich.console import Console
from rich.pretty import pprint

from truefoundry.deploy import Application, LocalSource
from truefoundry.pydantic_v1 import BaseModel


def generate_deployment_code(
    symbols_to_import: List[str],
    application_type: str,
    spec_repr: str,
    workspace_fqn: str,
):
    symbols = ", ".join(symbols_to_import)
    application_type = application_type.replace(" ", "").replace("-", "_")
    code = f"""\
import logging
from truefoundry.deploy import (
    {symbols},
)
logging.basicConfig(level=logging.INFO)

{application_type} = {spec_repr}

{application_type}.deploy(workspace_fqn="{workspace_fqn}", wait=False)\
"""
    return code


def extract_class_names(code):
    tree = ast.parse(code)

    # Function to extract keywords from the AST
    def extract_class_names_from_ast_tree(node):
        keywords = set()
        for child_node in ast.iter_child_nodes(node):
            if isinstance(child_node, ast.Call):
                keywords.add(child_node.func.id)
            keywords.update(extract_class_names_from_ast_tree(child_node))
        return keywords

    # Get keywords from the main body of the code
    main_keywords = extract_class_names_from_ast_tree(tree)
    return list(main_keywords)


def replace_enums_with_values(raw_str):
    # required to replace enums of format <AppProtocol.HTTP: 'http'> with 'http'
    pattern = r'<([a-zA-Z0-9_]+).[a-zA-Z0-9_]+: [\'"](.+)[\'"]>'
    replacement = r"'\2'"

    result = re.sub(pattern, replacement, raw_str)
    return result


def remove_none_type_fields(code):
    lines = code.split("\n")
    new_lines = [
        line
        for line in lines
        if not (line.endswith("=None") or line.endswith("=None,"))
    ]
    formatted_code = "\n".join(new_lines)
    return formatted_code


def remove_type_field(code):
    lines = code.split("\n")
    new_lines = [re.sub(r'^[ \t]*type=[\'"][^"]*[\'"],?', "", line) for line in lines]
    return "\n".join(new_lines)


def add_deploy_line(code, workspace_fqn, application_type):
    deploy_line = f"{application_type}.deploy('workspace_fqn={workspace_fqn}')"
    return code + "\n" + deploy_line


def get_python_repr(obj):
    stream = io.StringIO()
    console = Console(file=stream, no_color=True, highlighter=None)
    pprint(obj, expand_all=True, console=console, indent_guides=False)
    return stream.getvalue()


COMMENT_FOR_LOCAL_SOURCE = """# Set build_source=LocalSource(local_build=False), in order to deploy code from your local.
# With local_build=False flag, docker image will be built on cloud instead of local
# Else it will try to use docker installed on your local machine to build the image"""


def add_local_source_comment(code):
    lines = code.split("\n")
    new_lines = []
    for line in lines:
        if line.lstrip(" ").startswith("build_source=GitSource"):
            new_lines.append(COMMENT_FOR_LOCAL_SOURCE)
        new_lines.append(line)
    return "\n".join(new_lines)


def _convert_deployment_config_to_python(workspace_fqn: str, application_spec: dict):
    """
    Convert a deployment config to a python file that can be used to deploy to a workspace
    """
    application = Application.parse_obj(application_spec)
    application_obj = application.__root__
    application_type = application_obj.type
    if (
        hasattr(application_obj, "image")
        and hasattr(application_obj.image, "type")
        and application_obj.image.type == "build"
        and application_obj.image.build_source.type == "remote"
    ):
        application_obj.image.build_source = LocalSource(local_build=False)

    spec_repr = get_python_repr(application_obj)
    spec_repr = replace_enums_with_values(spec_repr)
    spec_repr = remove_none_type_fields(spec_repr)
    spec_repr = remove_type_field(spec_repr)

    # extract class names to import
    symbols_to_import = extract_class_names(spec_repr)

    # check if GitSource exists in array of symbols to import
    if "GitSource" in symbols_to_import:
        symbols_to_import.append("LocalSource")

    generated_code = generate_deployment_code(
        symbols_to_import=symbols_to_import,
        application_type=application_type,
        spec_repr=spec_repr,
        workspace_fqn=workspace_fqn,
    )

    if "GitSource" in symbols_to_import:
        generated_code = add_local_source_comment(generated_code)

    return generated_code


def _default_exclude_fn(model: BaseModel, name: str) -> bool:
    return False


def convert_deployment_config_to_python(
    workspace_fqn: str,
    application_spec: Dict[str, Any],
    exclude_fn: Callable[[BaseModel, str], bool] = _default_exclude_fn,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
):
    original_repr_args = BaseModel.__repr_args__

    def _patched_repr_args(self: BaseModel):
        _missing = object()
        pairs = []
        for name, value in original_repr_args(self):
            if name is not None:
                if exclude_fn(self, name):
                    continue
                if exclude_unset and name not in self.__fields_set__:
                    continue
                model_field = self.__fields__.get(name)
                if model_field is None:
                    continue
                if (
                    exclude_defaults
                    and not getattr(model_field, "required", True)
                    and getattr(model_field, "default", _missing) == value
                ):
                    continue
            pairs.append((name, value))
        return pairs

    try:
        BaseModel.__repr_args__ = _patched_repr_args
        return _convert_deployment_config_to_python(
            workspace_fqn=workspace_fqn, application_spec=application_spec
        )
    finally:
        BaseModel.__repr_args__ = original_repr_args


def generate_python_snippet_for_trigger_job(
    application_fqn: str,
    command: Optional[str],
    params: Optional[Dict[str, str]],
    run_name_alias: Optional[str],
):
    job_run_python_template = """\
from truefoundry.deploy import trigger_job

response = trigger_job(
  application_fqn="{{application_fqn}}",
  # You can pass command or params, but not both
  # command={{command}},
  # params={{params}},
  # run_name_alias={{run_name_alias}}
)

print(response.jobRunName)
"""
    output_python_str = job_run_python_template.replace(
        "{{application_fqn}}", application_fqn
    )

    if command is not None:
        output_python_str = output_python_str.replace("# command", "command")
        output_python_str = output_python_str.replace("{{command}}", repr(command))
    else:
        output_python_str = output_python_str.replace(
            "{{command}}", "<Enter Command Here>"
        )

    if params is not None:
        output_python_str = output_python_str.replace("# params", "params")
        output_python_str = output_python_str.replace("{{params}}", repr(params))
    else:
        output_python_str = output_python_str.replace(
            "{{params}}", "<Enter Params(key-value pairs) here as python dict>"
        )
    if run_name_alias is not None:
        output_python_str = output_python_str.replace(
            "# run_name_alias", "run_name_alias"
        )
        output_python_str = output_python_str.replace(
            "{{run_name_alias}}", repr(run_name_alias)
        )
    else:
        output_python_str = output_python_str.replace(
            "{{run_name_alias}}", "<Enter Run Name Alias here>"
        )

    return output_python_str


def generate_curl_snippet_for_trigger_job(
    control_plane_url: str,
    application_id: str,
    command: Optional[str],
    params: Optional[Dict[str, str]],
    run_name_alias: Optional[str],
):
    job_run_curl_request_template = """curl -X 'POST' \\
  '{{control_plane_url}}/api/svc/v1/jobs/trigger' \\
  -H 'accept: */*' \\
  -H 'Authorization: Bearer <Paste your API key here. You can generate it from the Settings Page>' \\
  -H 'Content-Type: application/json' \\
  -d '{
  "applicationId": "{{application_id}}",
  "input": {{input}},
  "metadata": {
    "job_run_name_alias": "{{run_name_alias}}"
  }
}'
"""
    output_curl_str = job_run_curl_request_template.replace(
        "{{control_plane_url}}", control_plane_url.rstrip("/")
    )
    output_curl_str = output_curl_str.replace("{{application_id}}", application_id)
    output_curl_str = output_curl_str.replace(
        "{{input}}", json.dumps({"command": command, "params": params}, indent=2)
    )
    if run_name_alias is not None:
        output_curl_str = output_curl_str.replace("{{run_name_alias}}", run_name_alias)
    else:
        output_curl_str = output_curl_str.replace(
            "{{run_name_alias}}", "<Enter Run Name Alias here>"
        )
    return output_curl_str



import json
from datetime import datetime, timezone
from typing import Any, Dict, Optional
from urllib.parse import urlparse

import rich_click as click

from truefoundry.cli.const import COMMAND_CLS
from truefoundry.cli.util import handle_exception_wrapper
from truefoundry.common.session import Session
from truefoundry.deploy.cli.commands.utils import (
    CONTEXT_NAME_FORMAT,
    get_cluster_server_url,
    get_kubeconfig_content,
    get_kubeconfig_path,
)
from truefoundry.deploy.io.no_output_callback import NoOutputCallBack
from truefoundry.deploy.lib.session import login


@click.command(
    name="k8s-exec-credential",
    cls=COMMAND_CLS,
    help="Generate a Kubernetes exec credential for the specified cluster user",
)
@click.option(
    "-c",
    "--cluster",
    type=str,
    required=True,
    help="The cluster id from TrueFoundry",
)
@handle_exception_wrapper
def k8s_exec_credential_command(cluster: str) -> None:
    """
    Generate a Kubernetes exec credential for the specified cluster.
    This command retrieves the cluster server URL from the kubeconfig file,
    """
    path = get_kubeconfig_path()
    kubeconfig: Dict[str, Any] = get_kubeconfig_content(path=path)
    server_url: Optional[str] = get_cluster_server_url(kubeconfig, cluster)
    if not server_url:
        raise click.ClickException(
            f"\nContext {CONTEXT_NAME_FORMAT.format(cluster=cluster)!r} for cluster {cluster!r} not found in kubeconfig. \n\nPlease run 'tfy get kubeconfig --cluster {cluster}' first."
        )
    host: str = f"{urlparse(server_url).scheme}://{urlparse(server_url).netloc}"
    login(host=host, output_hook=NoOutputCallBack())

    session = Session.new()
    token: str = session.access_token

    exec_credential: Dict[str, Any] = {
        "kind": "ExecCredential",
        "apiVersion": "client.authentication.k8s.io/v1beta1",
        "spec": {},
        "status": {
            "expirationTimestamp": datetime.fromtimestamp(
                session.token.exp, tz=timezone.utc
            ).strftime("%Y-%m-%dT%H:%M:%SZ"),
            "token": token,
        },
    }

    print(json.dumps(exec_credential, indent=4))



import rich_click as click

from truefoundry.ml import MlFoundryException
from truefoundry.ml.validation_utils import (
    _APP_NAME_REGEX,
)


class AppName(click.ParamType):
    """
    Custom ParamType to validate application names.
    """

    name = "application-name"

    def convert(self, value, param, ctx):
        try:
            if not value or not _APP_NAME_REGEX.match(value):
                raise MlFoundryException(
                    f"{value!r} must be lowercase and cannot contain spaces. It can only contain alphanumeric characters and hyphens. "
                    f"Length must be between 1 and 30 characters."
                )
        except MlFoundryException as e:
            self.fail(str(e), param, ctx)
        return value


class NonEmptyString(click.ParamType):
    name = "non-empty-string"

    def convert(self, value, param, ctx):
        if isinstance(value, str) and not value.strip():
            self.fail("Value cannot be empty or contain only spaces.", param, ctx)
        return value



import subprocess
from typing import Dict, List, Optional

import docker
from rich.console import Console

from truefoundry.deploy._autogen.models import BuildSecret
from truefoundry.logger import logger

__all__ = [
    "build_docker_image",
    "push_docker_image",
    "pull_docker_image",
    "push_docker_image_with_latest_tag",
    "env_has_docker",
]


def _get_docker_client():
    try:
        return docker.from_env()
    except Exception as ex:
        raise Exception("Could not connect to Docker") from ex


def env_has_docker():
    try:
        _get_docker_client()
        return True
    except Exception:
        return False


# this is required since push does throw an error if it
# fails - so we have to parse the response logs to catch the error
# the other option is to run `docker login` as a subprocess, but it's
# not recommended to provide password to subprocess
def _catch_error_in_push(response: List[dict]):
    for line in response:
        if line.get("error") is not None:
            raise Exception(
                f"Failed to push to registry with message '{line.get('error')}'"
            )


def _run_cmds(command: List[str]):
    console = Console(color_system=None, markup=False, soft_wrap=True)
    with subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    ) as sp:
        for line in sp.stdout:
            console.print(line.decode("utf-8").strip())
        sp.communicate()
        if sp.returncode is not None and sp.returncode != 0:
            raise Exception(f"Command: {command} failed")


def build_docker_image(
    path: str,
    tag: str,
    platform: str,
    dockerfile: str,
    extra_opts: Optional[List[str]] = None,
    build_args: Optional[Dict[str, str]] = None,
    build_secrets: Optional[List[BuildSecret]] = None,
):
    logger.info("Starting docker build...")
    try:
        # TODO (chiragjn): Maybe consider using client.images.build
        build_args_list = []
        if build_args:
            for k, v in build_args.items():
                build_args_list += ["--build-arg", f"{k}={v}"]

        docker_build_cmd = [
            "docker",
            "build",
            "-t",
            tag,
            "-f",
            dockerfile,
            "--platform",
            platform,
        ]

        build_secrets_list: List[str] = []
        if build_secrets:
            for build_secret in build_secrets:
                build_secrets_list.extend(
                    ["--secret", f"id={build_secret.id},env={build_secret.env}"]
                )
            docker_build_cmd += build_secrets_list
        docker_build_cmd += [path]
        docker_build_cmd += build_args_list
        docker_build_cmd += extra_opts if extra_opts else []
        _run_cmds(docker_build_cmd)
    except Exception as e:
        raise Exception(f"Error while building Docker image: {e}") from e


def push_docker_image(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    client = _get_docker_client()
    auth_config = {"username": docker_login_username, "password": docker_login_password}
    logger.info(f"Pushing {image_uri}")
    response = client.images.push(
        repository=image_uri, auth_config=auth_config, decode=True, stream=True
    )
    _catch_error_in_push(response=response)


def push_docker_image_with_latest_tag(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    client = _get_docker_client()
    auth_config = {"username": docker_login_username, "password": docker_login_password}
    repository_without_tag, _ = image_uri.rsplit(":", 1)
    image = client.images.get(image_uri)
    image.tag(repository=repository_without_tag, tag="latest")

    logger.info(f"Pushing {repository_without_tag}:latest")
    response = client.images.push(
        repository=repository_without_tag,
        tag="latest",
        auth_config=auth_config,
        decode=True,
        stream=True,
    )
    _catch_error_in_push(response=response)


def pull_docker_image(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    auth_config = {"username": docker_login_username, "password": docker_login_password}
    logger.info(f"Pulling cache image {image_uri}")
    _get_docker_client().images.pull(repository=image_uri, auth_config=auth_config)



import copy
import datetime
import json
import os
import tempfile
import uuid
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, NamedTuple, Optional, Union

from truefoundry import client
from truefoundry.common.warnings import TrueFoundryDeprecationWarning
from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    ArtifactDto,
    ArtifactManifest,
    ArtifactType,
    ArtifactVersionDto,
    CreateArtifactVersionRequestDto,
    DeleteArtifactVersionsRequestDto,
    ExternalBlobStorageSource,
    FinalizeArtifactVersionRequestDto,
    MlfoundryArtifactsApi,
    NotifyArtifactVersionFailureDto,
    TrueFoundryManagedSource,
    UpdateArtifactVersionRequestDto,
)
from truefoundry.ml.artifact.truefoundry_artifact_repo import (
    ArtifactIdentifier,
    MlFoundryArtifactsRepository,
)
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.log_types.artifacts.constants import (
    ARTIFACT_METADATA_TRUEFOUNDRY_KEY,
    INTERNAL_METADATA_PATH,
)
from truefoundry.ml.log_types.artifacts.utils import (
    _get_src_dest_pairs,
    _validate_artifact_metadata,
    _validate_description,
    calculate_total_size,
    get_autogen_type,
    set_tfy_internal_metadata,
    set_user_artifact_metadata,
)
from truefoundry.ml.logger import logger
from truefoundry.ml.session import _get_api_client
from truefoundry.pydantic_v1 import BaseModel, Extra, StrictStr

if TYPE_CHECKING:
    from truefoundry.ml.mlfoundry_run import MlFoundryRun


class ArtifactPath(NamedTuple):
    src: str
    dest: Optional[str] = None


class BlobStorageDirectory(BaseModel):
    uri: StrictStr


class ArtifactVersionInternalMetadata(BaseModel):
    class Config:
        extra = Extra.allow

    files_dir: str  # relative to root


class ArtifactVersionDownloadInfo(BaseModel):
    download_dir: str


class ArtifactVersion:
    def __init__(
        self,
        artifact_version: ArtifactVersionDto,
        artifact: ArtifactDto,
    ) -> None:
        self._api_client = _get_api_client()
        self._mlfoundry_artifacts_api = MlfoundryArtifactsApi(
            api_client=self._api_client
        )
        self._artifact_version: ArtifactVersionDto = artifact_version
        self._artifact: ArtifactDto = artifact
        self._deleted = False
        self._description: str = ""
        self._metadata: Dict[str, Any] = {}
        self._version_alias: Optional[str] = None
        self._set_mutable_attrs()

    @classmethod
    def from_fqn(cls, fqn: str) -> "ArtifactVersion":
        """
        Get the version of an Artifact to download contents or load them in memory

        Args:
            fqn (str): Fully qualified name of the artifact version.

        Returns:
            ArtifactVersion: An ArtifactVersion instance of the artifact

        Examples:

            ```python
            from truefoundry.ml import get_client, ArtifactVersion

            client = get_client()
            artifact_version = ArtifactVersion.from_fqn(fqn="<artifact-fqn>")
            ```
        """
        api_client = _get_api_client()
        mlfoundry_artifacts_api = MlfoundryArtifactsApi(api_client=api_client)
        _artifact_version = mlfoundry_artifacts_api.get_artifact_version_by_fqn_get(
            fqn=fqn
        )
        artifact_version = _artifact_version.artifact_version
        _artifact = mlfoundry_artifacts_api.get_artifact_by_id_get(
            id=artifact_version.artifact_id
        )
        return cls(
            artifact_version=_artifact_version.artifact_version,
            artifact=_artifact.artifact,
        )

    def _ensure_not_deleted(self):
        if self._deleted:
            raise MlFoundryException(
                "Artifact Version was deleted, cannot access a deleted version"
            )

    def _set_mutable_attrs(self):
        if self._artifact_version.manifest:
            manifest = self._artifact_version.manifest.actual_instance
            self._description = manifest.description or ""
            self._metadata = copy.deepcopy(manifest.metadata)
            self._version_alias = manifest.version_alias or None
        else:
            self._description = self._artifact_version.description or ""
            self._metadata = copy.deepcopy(
                self._artifact_version.artifact_metadata or {}
            )
            self._version_alias = None

    def _refetch_artifact_version(self, reset_mutable_attrs: bool = True):
        _artifact_version = (
            self._mlfoundry_artifacts_api.get_artifact_version_by_id_get(
                id=self._artifact_version.id
            )
        )
        self._artifact_version = _artifact_version.artifact_version
        if reset_mutable_attrs:
            self._set_mutable_attrs()

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r})"

    def _get_artifacts_repo(self):
        return MlFoundryArtifactsRepository(
            artifact_identifier=ArtifactIdentifier(
                artifact_version_id=uuid.UUID(self._artifact_version.id)
            ),
            api_client=self._api_client,
        )

    @property
    def name(self) -> str:
        """Get the name of the artifact"""
        return self._artifact.name

    @property
    def artifact_fqn(self) -> str:
        """Get fqn of the artifact"""
        return self._artifact.fqn

    @property
    def version(self) -> int:
        """Get version information of the artifact"""
        return self._artifact_version.version

    @property
    def fqn(self) -> str:
        """Get fqn of the current artifact version"""
        return self._artifact_version.fqn

    @property
    def step(self) -> Optional[int]:
        """Get the step in which artifact was created"""
        if self._artifact_version.manifest:
            return self._artifact_version.manifest.actual_instance.step
        return self._artifact_version.step

    @property
    def description(self) -> str:
        """Get description of the artifact"""
        return self._description

    @description.setter
    def description(self, value: str):
        """set the description of the artifact"""
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        """Get metadata for the current artifact"""
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        """set the metadata for current artifact"""
        _validate_artifact_metadata(value)
        self._metadata = copy.deepcopy(value)

    @property
    def version_alias(self) -> Optional[str]:
        """
        Get version alias for the current artifact version
        """
        if not self._artifact_version.manifest:
            warnings.warn(
                message="This artifact version was created using an older serialization format. version alias does not exist",
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )
        return self._version_alias

    @version_alias.setter
    def version_alias(self, value: Optional[str]):
        """
        Set the version alias for current artifact version
        """
        if not self._artifact_version.manifest:
            warnings.warn(
                message="This artifact version was created using an older serialization format. version alias will not be updated",
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )
            return
        self._version_alias = value

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        """Get the time at which artifact was created"""
        return self._artifact_version.created_at

    @property
    def updated_at(self) -> Optional[datetime.datetime]:
        """Get the information about when the artifact was updated"""
        return self._artifact_version.updated_at

    def raw_download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> str:
        """
        Download an artifact file or directory to a local directory if applicable, and return a
        local path for it.

        Args:
            path (str): Absolute path of the local filesystem destination directory to which to
                        download the specified artifacts. This directory must already exist.
                        If unspecified, the artifacts will either be downloaded to a new
                        uniquely-named directory on the local filesystem.
            overwrite (bool): If True it will overwrite the file if it is already present in the download directory else
                              it will throw an error
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            path:  Absolute path of the local filesystem location containing the desired artifacts.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="<your-artifact-fqn>")
            artifact_version.raw_download(path="<your-desired-download-path>")
            ```
        """
        logger.info(
            "Downloading artifact version contents, this might take a while ..."
        )
        artifacts_repo = self._get_artifacts_repo()
        return artifacts_repo.download_artifacts(
            artifact_path="", dst_path=path, overwrite=overwrite, progress=progress
        )

    def _download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> ArtifactVersionDownloadInfo:
        self._ensure_not_deleted()
        download_dir = self.raw_download(
            path=path, overwrite=overwrite, progress=progress
        )
        if self._artifact_version.manifest:
            download_info = ArtifactVersionDownloadInfo(
                download_dir=download_dir,
            )
            return download_info

        internal_metadata_path = os.path.join(download_dir, INTERNAL_METADATA_PATH)
        if not os.path.exists(internal_metadata_path):
            raise MlFoundryException(
                "Artifact version seems to be corrupted or in invalid format due to missing artifact metadata. "
                "You can still use .raw_download(path='/your/path/here') to download and inspect files."
            )
        with open(internal_metadata_path) as f:
            internal_metadata = ArtifactVersionInternalMetadata.parse_obj(json.load(f))
        download_path = os.path.join(download_dir, internal_metadata.files_dir)
        return ArtifactVersionDownloadInfo(download_dir=download_path)

    def download(
        self,
        path: Optional[Union[str, Path]] = None,
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> str:
        """
        Download an artifact file or directory to a local directory if applicable, and return a
        local path for it.

        Args:
            path (str): Absolute path of the local filesystem destination directory to which to
                        download the specified artifacts. This directory must already exist.
                        If unspecified, the artifacts will either be downloaded to a new
                        uniquely-named directory on the local filesystem or will be returned
                        directly in the case of the Local ArtifactRepository.
            overwrite (bool): If True it will overwrite the file if it is already present in the download directory else
                              it will throw an error
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            path:  Absolute path of the local filesystem location containing the desired artifacts.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="<your-artifact-fqn>")
            artifact_version.download(path="<your-desired-download-path>")
            ```
        """
        download_info = self._download(
            path=path, overwrite=overwrite, progress=progress
        )
        logger.info("Downloaded artifact contents to %s", download_info.download_dir)
        return download_info.download_dir

    def delete(self) -> bool:
        """
        Deletes the current instance of the ArtifactVersion hence deleting the current version.

        Returns:
            True if artifact was deleted successfully

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="<your-artifact-fqn>")
            artifact_version.delete()
            ```
        """
        self._ensure_not_deleted()
        self._mlfoundry_artifacts_api.delete_artifact_version_post(
            delete_artifact_versions_request_dto=DeleteArtifactVersionsRequestDto(
                id=self._artifact_version.id
            )
        )
        self._deleted = True
        return True

    def update(self):
        """
        Updates the current instance of the ArtifactVersion hence updating the current version.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="<your-artifact-fqn>")
            artifact_version.description = 'This is the new description'
            artifact_version.update()
            ```
        """
        self._ensure_not_deleted()
        if self._artifact_version.manifest:
            manifest = self._artifact_version.manifest.actual_instance
            assert isinstance(manifest, ArtifactManifest)
            manifest.description = self.description
            manifest.metadata = self.metadata
            manifest.version_alias = self.version_alias
        else:
            manifest = None
        try:
            _manifest_cls = get_autogen_type(
                UpdateArtifactVersionRequestDto, "manifest"
            )
            _artifact_version = (
                self._mlfoundry_artifacts_api.update_artifact_version_post(
                    update_artifact_version_request_dto=UpdateArtifactVersionRequestDto(
                        id=self._artifact_version.id,
                        description=self.description,
                        artifact_metadata=self.metadata,
                        manifest=_manifest_cls.from_dict(manifest.to_dict())
                        if manifest
                        else None,
                    )
                )
            )
        except Exception:
            # rollback edits to internal object
            self._refetch_artifact_version(reset_mutable_attrs=False)
            raise
        else:
            self._artifact_version = _artifact_version.artifact_version
            self._set_mutable_attrs()


def _log_artifact_version_helper(
    run: Optional["MlFoundryRun"],
    name: str,
    artifact_type: ArtifactType,
    artifact_dir: Union[tempfile.TemporaryDirectory, BlobStorageDirectory],
    dest_to_src_map: Dict[str, str],
    ml_repo: Optional[str] = None,
    description: Optional[str] = None,
    internal_metadata: Optional[BaseModel] = None,
    metadata: Optional[Dict[str, Any]] = None,
    step: int = 0,
    progress: Optional[bool] = None,
) -> ArtifactVersion:
    if not run and not ml_repo:
        raise MlFoundryException("Exactly one of run, ml_repo should be passed")

    if run:
        mlfoundry_artifacts_api = run._mlfoundry_artifacts_api
        repos = client.ml_repos.get(id=run._experiment_id)
        ml_repo = repos.data.manifest.name
        ml_repo_id = run._experiment_id
    else:
        assert ml_repo is not None
        api_client = _get_api_client()
        mlfoundry_artifacts_api = MlfoundryArtifactsApi(api_client=api_client)
        result = list(client.ml_repos.list(name=ml_repo, limit=1))[0]
        ml_repo_id = result.id

    metadata = metadata or {}
    metadata = set_tfy_internal_metadata(metadata)
    metadata = set_user_artifact_metadata(metadata)

    _validate_description(description)
    _validate_artifact_metadata(metadata)

    assert mlfoundry_artifacts_api is not None
    _create_artifact_response = mlfoundry_artifacts_api.create_artifact_version_post(
        create_artifact_version_request_dto=CreateArtifactVersionRequestDto(
            experiment_id=int(ml_repo_id),
            name=name,
            artifact_type=artifact_type,
        )
    )
    version_id = _create_artifact_response.id
    artifact_storage_root = _create_artifact_response.artifact_storage_root
    total_size = None

    if isinstance(artifact_dir, tempfile.TemporaryDirectory):
        # Source is of type TrueFoundryManagedSource
        source = TrueFoundryManagedSource(type="truefoundry", uri=artifact_storage_root)
        artifacts_repo = MlFoundryArtifactsRepository(
            artifact_identifier=ArtifactIdentifier(
                artifact_version_id=uuid.UUID(version_id),
            ),
            api_client=mlfoundry_artifacts_api.api_client,
        )

        total_size = calculate_total_size(list(dest_to_src_map.values()))
        try:
            logger.info(
                "Packaging and uploading files to remote with size: %.6f MB",
                total_size / 1000000.0,
            )
            src_dest_pairs = _get_src_dest_pairs(
                root_dir=artifact_dir.name, dest_to_src_map=dest_to_src_map
            )
            artifacts_repo.log_artifacts(
                src_dest_pairs=src_dest_pairs, progress=progress
            )
        except Exception as e:
            mlfoundry_artifacts_api.notify_failure_post(
                notify_artifact_version_failure_dto=NotifyArtifactVersionFailureDto(
                    id=version_id
                )
            )
            raise MlFoundryException("Failed to log Artifact") from e
        finally:
            artifact_dir.cleanup()
    elif isinstance(artifact_dir, BlobStorageDirectory):
        source = ExternalBlobStorageSource(type="external", uri=artifact_dir.uri)
    else:
        raise MlFoundryException("Invalid artifact_dir provided")

    if total_size is not None and total_size > 0:
        metadata[ARTIFACT_METADATA_TRUEFOUNDRY_KEY]["artifact_size"] = total_size

    artifact_manifest = None
    if artifact_type == ArtifactType.ARTIFACT:
        _source_cls = get_autogen_type(ArtifactManifest, "source")
        artifact_manifest = ArtifactManifest(
            name=name,
            ml_repo=ml_repo,
            description=description,
            metadata=metadata,
            source=_source_cls.from_dict(source.dict()),
            step=step,
            run_id=run.run_id if run else None,
        )
    _manifest_cls = get_autogen_type(FinalizeArtifactVersionRequestDto, "manifest")

    finalize_artifact_version_request_dto = FinalizeArtifactVersionRequestDto(
        id=version_id,
        run_uuid=run.run_id if run else None,
        internal_metadata=internal_metadata.dict()
        if internal_metadata is not None
        else {},
        artifact_metadata=metadata,
        data_path=INTERNAL_METADATA_PATH if internal_metadata else None,
        step=step,
        artifact_size=total_size,
        manifest=_manifest_cls.from_dict(artifact_manifest.to_dict())
        if artifact_manifest
        else None,
    )
    _artifact_version = mlfoundry_artifacts_api.finalize_artifact_version_post(
        finalize_artifact_version_request_dto=finalize_artifact_version_request_dto
    )
    return ArtifactVersion.from_fqn(fqn=_artifact_version.artifact_version.fqn)



from __future__ import annotations

import json
from typing import ClassVar, Iterable, List, Protocol, Type, runtime_checkable

from openai import OpenAI
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.chat.chat_completion_system_message_param import (
    ChatCompletionSystemMessageParam,
)
from openai.types.chat.chat_completion_tool_message_param import (
    ChatCompletionToolMessageParam,
)
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from openai.types.chat.chat_completion_user_message_param import (
    ChatCompletionUserMessageParam,
)
from openai.types.shared_params.function_definition import FunctionDefinition

from truefoundry.autodeploy.constants import AUTODEPLOY_MODEL_NAME
from truefoundry.autodeploy.logger import logger
from truefoundry.autodeploy.tools import Event, RequestEvent, Tool
from truefoundry.autodeploy.utils.pydantic_compat import model_dump, model_json_schema
from truefoundry.pydantic_v1 import BaseModel, ValidationError


def llm(
    openai_client: OpenAI,
    messages: List[ChatCompletionMessageParam],
    tools: List[ChatCompletionToolParam],
    model: str,
    max_tokens: int = 4096,
) -> ChatCompletionMessage:
    completion = openai_client.chat.completions.create(
        tools=tools,
        stream=False,
        messages=messages,
        model=model,
        max_tokens=max_tokens,
        temperature=1,
        top_p=0.01,
        n=1,
        frequency_penalty=0.01,
    )
    return completion.choices[0].message


def format_tool_response(
    response: BaseModel, tool_call_id: str
) -> ChatCompletionToolMessageParam:
    return ChatCompletionToolMessageParam(
        role="tool",
        content=json.dumps(model_dump(response), indent=1),
        tool_call_id=tool_call_id,
    )


def format_user_response(response: str) -> ChatCompletionUserMessageParam:
    return ChatCompletionUserMessageParam(
        role="user",
        content=response,
    )


def get_tool_descriptions(
    tools: List, response: Type[BaseModel]
) -> List[ChatCompletionToolParam]:
    descriptions = []
    for tool in tools:
        tool = ChatCompletionToolParam(
            type="function",
            function=FunctionDefinition(
                name=tool.__class__.__name__,
                description=tool.description.strip(),
                parameters=model_json_schema(tool.Request),
            ),
        )
        descriptions.append(tool)

    if response:
        tool = ChatCompletionToolParam(
            type="function",
            function=FunctionDefinition(
                name="Response",
                parameters=model_json_schema(response),
            ),
        )
        descriptions.append(tool)

    return descriptions


class ToolParamParseError(BaseModel):
    error: str


@runtime_checkable
class Agent(Tool, Protocol):
    system_prompt: ClassVar[str]
    tools: List[Tool]
    max_iter: ClassVar[int] = 30
    openai_client: OpenAI
    model: str = AUTODEPLOY_MODEL_NAME

    def run(self, request: RequestEvent) -> Iterable[Event]:  # noqa: C901
        messages: List[ChatCompletionMessageParam] = [
            ChatCompletionSystemMessageParam(
                role="system",
                content=self.system_prompt + "\nrequest:\n" + request.json(),
            ),
        ]
        tool_descriptions = get_tool_descriptions(self.tools, self.Response)
        tool_map = {}
        for tool in self.tools:
            tool_map[tool.__class__.__name__] = tool
        for _ in range(self.max_iter):
            r = llm(
                messages=messages,
                tools=tool_descriptions,
                openai_client=self.openai_client,
                model=self.model,
            )
            messages.append(r)
            if r.content:
                logger.debug(r.content)
            if not r.tool_calls:
                messages.append(
                    format_user_response(
                        "You must respond with a tool call. Use the Ask tool to ask user."
                    )
                )
                continue

            for tool_call in r.tool_calls:
                if tool_call.function.name == "Response":
                    try:
                        response = self.Response(
                            **json.loads(tool_call.function.arguments)
                        )
                    except (json.decoder.JSONDecodeError, ValidationError) as ex:
                        logger.debug(f"{tool_call.function.arguments}, {ex}")
                        messages.append(
                            format_tool_response(
                                ToolParamParseError(error=str(ex)), tool_call.id
                            )
                        )
                        continue
                    logger.debug(response)
                    yield response
                    return response
                tool = tool_map[tool_call.function.name]
                try:
                    request = tool.Request(**json.loads(tool_call.function.arguments))
                except (json.decoder.JSONDecodeError, ValidationError) as ex:
                    logger.debug(f"{tool_call.function.arguments}, {ex}")
                    messages.append(
                        format_tool_response(
                            ToolParamParseError(error=str(ex)), tool_call.id
                        )
                    )
                    continue
                logger.debug(f"{self.__class__.__name__}, {tool_call.function.name}")
                logger.debug(request)
                yield request
                tool_run = tool.run(request)
                response = None
                inp = None
                while True:
                    try:
                        event = tool_run.send(inp)
                        inp = yield event
                    except StopIteration as ex:
                        response = ex.value
                        break
                logger.debug(response)
                logger.debug(f"{self.__class__.__name__}, {tool_call.function.name}")
                messages.append(format_tool_response(response, tool_call.id))
                if not isinstance(tool, Agent):
                    yield response
        raise Exception()



import json
import logging
import os
import posixpath
import typing
from pathlib import Path, PureWindowsPath
from typing import Any, Dict, List, Optional, Sequence, Tuple, Type, Union

from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.log_types.artifacts.constants import (
    ARTIFACT_METADATA_TRUEFOUNDRY_KEY,
    DESCRIPTION_MAX_LENGTH,
    TFY_ARTIFACTS_ADDITIONAL_USER_METADATA_ENV_VAR,
    TFY_INTERNAL_APPLICATION_ID_ENV_VAR,
    TFY_INTERNAL_APPLICATION_VERSION_ENV_VAR,
    TFY_INTERNAL_JOB_RUN_NAME_ENV_VAR,
)
from truefoundry.pydantic_v1 import BaseModel

logger = logging.getLogger(__name__)


def to_unix_path(path):
    path = os.path.normpath(path)
    if os.path.sep == "\\":
        path = PureWindowsPath(path).as_posix()
    return path


def _copy_tree(
    root_dir: str, src_path: str, dest_path: str, dest_to_src: Dict[str, str]
):
    os.makedirs(dest_path, exist_ok=True)
    for item in os.listdir(src_path):
        src = os.path.join(src_path, item)
        dest = os.path.join(dest_path, item)
        if os.path.isdir(src):
            _copy_tree(
                root_dir=root_dir,
                src_path=src,
                dest_path=dest,
                dest_to_src=dest_to_src,
            )
        else:
            dest_to_src[dest] = src


def is_destination_path_dirlike(dest_path) -> bool:
    if not dest_path:
        return True

    if dest_path.endswith(os.sep) or dest_path.endswith(posixpath.sep):
        return True

    if os.path.exists(dest_path) and os.path.isdir(dest_path):
        return True

    return False


def get_single_file_path_if_only_one_in_directory(path: str) -> Optional[str]:
    """
    Get the filename from a path, or return a filename from a directory if a single file exists.
    Args:
        path: The file or folder path.
    Returns:
        Optional[str]: The filename or None if no files are found or multiple files are found.
    """
    # If it's already a file, return it as-is
    if os.path.isfile(path):
        return path

    # If it's a directory, check if it contains a single file
    if is_destination_path_dirlike(path):
        all_files: List[str] = []
        for root, _, files in os.walk(path):
            # Collect all files found in any subdirectory
            all_files.extend(os.path.join(root, f) for f in files)
            # If more than one file is found, stop early
            if len(all_files) > 1:
                return None

        # If only one file is found, return it
        if len(all_files) == 1:
            return all_files[0]

    return None  # No file found or Multiple files found


def _copy_additional_files(
    root_dir: str,
    files_dir: str,  # relative to root dir e.g. "files/"
    model_dir: Optional[str],  # relative to files_dir e.g "model/"
    additional_files: Sequence[Tuple[Union[str, Path], Optional[str]]],
    ignore_model_dir_dest_conflict: bool = False,
    existing_dest_to_src_map: Optional[Dict[str, str]] = None,
) -> Dict[str, str]:
    """
    File copying examples:
        # non ambiguous
        # a.txt -> /tmp/                            result /tmp/a.txt
        # a.txt -> /tmp/a/                          result /tmp/a/a.txt
        # a.txt -> /tmp/a/b/c/d.txt                 result /tmp/a/b/c/d.txt
        # .gitignore -> /tmp/.gitignore             result /tmp/.gitignore

        # ambiguous but destination directory exists
        # a.txt -> /tmp                             result /tmp/a.txt
        # a.txt -> /tmp/a (and /tmp/a/ exists)      result /tmp/a/a.txt

        # ambiguous - when the destination can't be reliably distinguished as a directory
        # a -> /tmp/a                                result /tmp/a
        # a -> /tmp/b                                result /tmp/b
        # a -> /tmp/a.txt                            result /tmp/a.txt
        # .gitignore -> /tmp/.gitinclude             result /tmp/.gitinclude
        # a.txt -> /tmp/a                            result /tmp/a
    """
    dest_to_src = existing_dest_to_src_map or {}
    for src_path, dest_path in additional_files:
        src_path = str(src_path)
        if not os.path.exists(src_path):
            raise MlFoundryException(
                f"Source path {src_path!r} in `additional_files` does not exist."
            )
        dest_path = dest_path or ""
        normalized_path = os.path.normpath(dest_path)
        if dest_path.endswith(os.sep) or dest_path.endswith(posixpath.sep):
            normalized_path += os.sep
        dest_path = normalized_path.lstrip(os.sep)

        if (
            model_dir
            and dest_path.startswith(model_dir)
            and not ignore_model_dir_dest_conflict
        ):
            logger.warning(
                f"Destination path {dest_path!r} in `additional_files` conflicts with "
                f"reserved path {model_dir!r}/ which is being used to store the model. "
                f"This might cause errors"
            )

        files_abs_dir = os.path.join(root_dir, files_dir)
        dest_abs_path = os.path.join(files_abs_dir, dest_path)

        if os.path.isfile(src_path):
            _src = src_path
            if is_destination_path_dirlike(dest_abs_path):
                os.makedirs(dest_abs_path, exist_ok=True)
                dest_abs_path = os.path.join(dest_abs_path, os.path.basename(_src))
            else:
                os.makedirs(os.path.dirname(dest_abs_path), exist_ok=True)
            _dst = os.path.relpath(dest_abs_path, files_abs_dir)
            logger.info(f"Adding file {_src} as /{_dst}")
            dest_to_src[dest_abs_path] = src_path
        elif os.path.isdir(src_path):
            os.makedirs(dest_abs_path, exist_ok=True)
            _src = src_path.rstrip("/")
            _dst = os.path.relpath(dest_abs_path, files_abs_dir).rstrip("/")
            logger.info(f"Adding contents of {_src}/ to /{_dst}/")
            _copy_tree(
                root_dir=root_dir,
                src_path=src_path,
                dest_path=dest_abs_path,
                dest_to_src=dest_to_src,
            )

    return dest_to_src


def _make_dest_to_src_map_from_dir(root_dir: str) -> Dict[str, str]:
    dest_to_src_map = {}
    for root, _, files in os.walk(root_dir):
        for file in files:
            src = os.path.join(root, file)
            dest = src
            dest_to_src_map[dest] = src
    return dest_to_src_map


def _get_src_dest_pairs(
    root_dir: str,
    dest_to_src_map: Dict[str, str],
) -> Sequence[Tuple[str, str]]:
    src_dest_pairs = [
        (src_path, to_unix_path(os.path.relpath(dest_abs_path, root_dir)))
        for dest_abs_path, src_path in dest_to_src_map.items()
    ]
    return src_dest_pairs


def _validate_description(description: Optional[str]):
    if description is not None:
        if not isinstance(description, str):
            raise MlFoundryException(
                "`description` must be either `None` or type `str`"
            )
        if len(description) > DESCRIPTION_MAX_LENGTH:
            raise MlFoundryException(
                f"`description` cannot be longer than {DESCRIPTION_MAX_LENGTH} characters"
            )


def _validate_artifact_metadata(metadata: Dict[str, Any]):
    if not isinstance(metadata, dict):
        raise MlFoundryException("`metadata` must be json serializable dict")
    try:
        json.dumps(metadata)
    except ValueError as ve:
        raise MlFoundryException("`metadata` must be json serializable dict") from ve


def calculate_total_size(
    paths: Sequence[str],
):
    """
    Tells about the size of the artifact

    Args:
        paths (str): list of paths to include in total size calculation

    Returns:
        total size of the artifact
    """
    return sum(os.stat(os.path.realpath(file_path)).st_size for file_path in paths)


def get_autogen_type(parent_type: Type[BaseModel], field_name: str) -> Type[BaseModel]:
    type_ = typing.get_type_hints(parent_type)[field_name]
    origin = typing.get_origin(type_)

    if origin is None:
        return type_
    elif origin is typing.Union:
        args = typing.get_args(type_)
        if len(args) == 2 and args[-1] is type(None):
            return args[0]

    raise NotImplementedError(f"Cannot extract main type from {type_}")


def set_tfy_internal_metadata(metadata: Dict[str, Any]) -> Dict[str, Any]:
    tfy_metadata = metadata.setdefault(ARTIFACT_METADATA_TRUEFOUNDRY_KEY, {})
    if not isinstance(tfy_metadata, dict):
        return metadata

    tfy_internal_metadata = {
        "application_id": os.environ.get(TFY_INTERNAL_APPLICATION_ID_ENV_VAR),
        "application_version": os.environ.get(TFY_INTERNAL_APPLICATION_VERSION_ENV_VAR),
        "job_run_name": os.environ.get(TFY_INTERNAL_JOB_RUN_NAME_ENV_VAR),
    }
    tfy_internal_metadata = {
        k: v for k, v in tfy_internal_metadata.items() if v is not None
    }

    if tfy_internal_metadata:
        created_by_key = "created_by"
        created_by_metadata = tfy_metadata.setdefault(created_by_key, {})
        if created_by_key in tfy_metadata and not isinstance(created_by_metadata, dict):
            return metadata
        for key, value in tfy_internal_metadata.items():
            if key not in created_by_metadata:
                created_by_metadata[key] = value
    return metadata


def _merge_dicts_recursively(
    src: Dict[str, Any], dest: Dict[str, Any], overwrite: bool = False
) -> Dict[str, Any]:
    for key, value in src.items():
        if key not in dest:
            dest[key] = value
        else:
            if isinstance(value, dict) and isinstance(dest[key], dict):
                _merge_dicts_recursively(value, dest[key], overwrite=overwrite)
            elif overwrite:
                dest[key] = value
    return dest


def set_user_artifact_metadata(metadata: Dict[str, Any]) -> Dict[str, Any]:
    _user_metadata = os.environ.get(TFY_ARTIFACTS_ADDITIONAL_USER_METADATA_ENV_VAR)
    if not _user_metadata:
        return metadata

    try:
        user_metadata = json.loads(_user_metadata)
    except ValueError:
        logger.warning(
            "Content of `TFY_ARTIFACTS_ADDITIONAL_USER_METADATA` environment variable is not valid json, cannot add to metadata"
        )
        return metadata

    if not isinstance(user_metadata, dict):
        logger.warning(
            "Content of `TFY_ARTIFACTS_ADDITIONAL_USER_METADATA` environment variable is not a dictionary, cannot add to metadata"
        )
        return metadata

    metadata = _merge_dicts_recursively(user_metadata, metadata, overwrite=False)
    return metadata



from typing import Optional

from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)
from truefoundry.deploy.lib.model.entity import Workspace


def list_workspaces(
    cluster_name: Optional[str] = None,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    workspaces = client.list_workspaces(cluster_id=cluster_name)
    return workspaces


def get_workspace_by_fqn(
    workspace_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient()
    workspaces = client.get_workspace_by_fqn(workspace_fqn=workspace_fqn)
    if len(workspaces) == 0:
        raise ValueError(f"Workspace with FQN {workspace_fqn!r} does not exist.")
    workspace = workspaces[0]
    return workspace


def delete_workspace(
    workspace_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient()
    workspace = client.get_id_from_fqn(fqn=workspace_fqn, fqn_type="workspace")
    deleted_workspace = client.delete_workspace(workspace_id=workspace["workspaceId"])
    return deleted_workspace



from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional

import yaml

from truefoundry.cli.console import console
from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)
from truefoundry.deploy.lib.messages import PROMPT_DELETING_MANIFEST
from truefoundry.deploy.lib.model.entity import DeleteResult, File, ManifestLike
from truefoundry.pydantic_v1 import ValidationError


def _delete_manifest(
    manifest: Dict[str, Any],
    client: Optional[ServiceFoundryServiceClient] = None,
    filename: Optional[str] = None,
    index: Optional[int] = None,
) -> DeleteResult:
    client = client or ServiceFoundryServiceClient()

    file_metadata = ""
    if index is not None:
        file_metadata += f" at index {index}"
    if filename:
        file_metadata += f" from file {filename}"

    try:
        parsed_manifest = ManifestLike.parse_obj(manifest)
    except ValidationError as ex:
        return DeleteResult(
            success=False,
            message=f"Failed to parse manifest{file_metadata}. Error: {ex}",
        )

    manifest_name = ""
    if parsed_manifest.name:
        manifest_name = f" {parsed_manifest.name}"

    try:
        client.delete(parsed_manifest.dict(exclude_unset=True, exclude_none=True))

        return DeleteResult(
            success=True,
            message=(
                f"Successfully deleted resource manifest{manifest_name} of type {parsed_manifest.type}."
            ),
        )
    except Exception as ex:
        return DeleteResult(
            success=False,
            message=(
                f"Failed to delete resource manifest{manifest_name} of type {parsed_manifest.type}. Error: {ex}."
            ),
        )


def delete_manifest(
    manifest: Dict[str, Any],
    client: Optional[ServiceFoundryServiceClient] = None,
) -> DeleteResult:
    return _delete_manifest(manifest=manifest, client=client)


def delete_manifest_resources(
    filename: str,
    manifests_it: List[Dict[str, Any]],
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Iterator[DeleteResult]:
    for index, manifest in enumerate(manifests_it):
        if not isinstance(manifest, dict):
            yield DeleteResult(
                success=False,
                message=f"Failed to delete resource manifest at index {index} from file {filename}. Error: A manifest must be a dict, got type {type(manifest)}",
            )
            continue

        yield _delete_manifest(
            manifest=manifest,
            client=client,
            filename=filename,
            index=index,
        )


def delete_manifest_file(
    filepath: str,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Iterator[DeleteResult]:
    client = client or ServiceFoundryServiceClient()
    filename = Path(filepath).name
    try:
        with open(filepath, "r") as f:
            manifests_it = list(yaml.safe_load_all(f))
    except Exception as ex:
        yield DeleteResult(
            success=False,
            message=f"Failed to read file {filepath} as a valid YAML file. Error: {ex}",
        )
    else:
        yield from delete_manifest_resources(
            filename=filename, manifests_it=manifests_it, client=client
        )


def delete_manifest_files(
    files: List[File],
    sf_server_client: Optional[ServiceFoundryServiceClient] = None,
    read_content_from_file: bool = True,
) -> List[DeleteResult]:
    delete_results: List[DeleteResult] = []
    sf_server_client = sf_server_client or ServiceFoundryServiceClient()
    for file_info in files:
        file_path = file_info.path
        delete_result_it: Iterator[DeleteResult] = []
        if read_content_from_file:
            delete_result_it = delete_manifest_file(file_path, sf_server_client)
        else:
            try:
                manifests_it = [yaml.safe_load(file_info.content)]
            except Exception:
                continue
            delete_result_it = delete_manifest_resources(
                file_path, manifests_it, sf_server_client
            )

        with console.status(PROMPT_DELETING_MANIFEST.format(file_path), spinner="dots"):
            for delete_result in delete_result_it:
                if delete_result.success:
                    console.print(f"[green]\u2714 {delete_result.message}[/]")
                else:
                    console.print(f"[red]\u2718 {delete_result.message}[/]")

                delete_results.append(delete_result)
    return delete_results



from flytekit import ContainerTask

from truefoundry.workflow import ContainerTaskConfig


class ContainerTask(ContainerTask):
    def __init__(self, name: str, task_config: ContainerTaskConfig):
        super().__init__(name=name, image="", command=[])
        self.tfy_task_config = task_config

    def get_custom(self, settings) -> dict:
        return {"truefoundry": self.tfy_task_config.dict()}



from typing import Any, Dict, Optional
from urllib.parse import urljoin

import rich_click as click
from rich.console import Console

from truefoundry.cli.const import COMMAND_CLS
from truefoundry.cli.util import handle_exception_wrapper, select_cluster
from truefoundry.common.session import Session
from truefoundry.deploy.cli.commands.utils import (
    CONTEXT_NAME_FORMAT,
    add_update_cluster_context,
    get_cluster_server_url,
    get_kubeconfig_content,
    get_kubeconfig_path,
    save_kubeconfig,
)

console = Console()


def _construct_k8s_proxy_server(host: str, cluster: str) -> str:
    """
    Construct the Kubernetes proxy server URL.
    """
    return urljoin(host, f"api/svc/v1/k8s/proxy/{cluster}")


def _should_update_existing_context(cluster: str, kubeconfig: Dict[str, Any]) -> bool:
    """
    Prompt the user whether to overwrite an existing kubeconfig context.
    """
    server_url = get_cluster_server_url(kubeconfig, cluster)
    if server_url is not None:
        console.print(
            f"\nContext {CONTEXT_NAME_FORMAT.format(cluster=cluster)!r} for cluster {cluster!r} already exists in kubeconfig.\n"
        )
        return click.confirm(
            text="Do you want to update the context?", default=False, err=True
        )
    return True


@click.command(name="kubeconfig", cls=COMMAND_CLS)
@click.option(
    "-c",
    "--cluster",
    type=str,
    required=False,
    help="The cluster id from TrueFoundry. If not provided, an interactive prompt will list available clusters",
)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    show_default=True,
    help="Overwrites existing cluster entry without prompting",
)
@handle_exception_wrapper
def kubeconfig_command(cluster: Optional[str] = None, overwrite: bool = False) -> None:
    """
    Update kubeconfig file to access cluster attached to TrueFoundry Control Plane.

    By default, credentials are written to ~/.kube/config. You can provide an alternate path by setting the KUBECONFIG environment variable. If KUBECONFIG contains multiple paths, the first one is used.
    """
    session = Session.new()
    cluster = select_cluster(cluster)

    path = get_kubeconfig_path()
    kubeconfig = get_kubeconfig_content(path=path)

    if not overwrite and not _should_update_existing_context(cluster, kubeconfig):
        console.print(
            "Existing context found. Use '--overwrite' to force update the context."
        )
        return

    k8s_proxy_server = _construct_k8s_proxy_server(session.tfy_host, cluster)
    context_name = add_update_cluster_context(
        kubeconfig,
        cluster,
        k8s_proxy_server,
        exec_command=[
            "tfy",
            "--json",
            "get",
            "k8s-exec-credential",
            "--cluster",
            cluster,
        ],
        envs={"TFY_INTERNAL": "1"},
    )

    save_kubeconfig(kubeconfig, path=path)
    console.print(
        f"\nUpdated kubeconfig at {str(path)!r} with context {context_name!r} for cluster {cluster!r}\n\n"
        f"Run 'kubectl config use-context {context_name}' to use this context.\n"
    )



import logging

from truefoundry.common.constants import ENV_VARS
from truefoundry.workflow.remote_filesystem.logger import init_logger

init_logger(level=ENV_VARS.TFY_INTERNAL_SIGNED_URL_CLIENT_LOG_LEVEL)



import os
import sys
from typing import Optional

import rich_click as click
import yaml
from click import UsageError
from click.exceptions import ClickException

from truefoundry.autodeploy.cli import cli as autodeploy_cli
from truefoundry.autodeploy.exception import InvalidRequirementsException
from truefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from truefoundry.cli.util import handle_exception_wrapper


def _get_default_spec_file():
    paths = [
        "./truefoundry.yaml",
        "./truefoundry.yml",
        "./servicefoundry.yaml",
        "./servicefoundry.yml",
    ]
    for path in paths:
        if os.path.exists(path):
            return path


@click.group(
    name="deploy",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="Deploy application to TrueFoundry",
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    default=_get_default_spec_file(),
    help="Path to truefoundry.yaml file",
    show_default=True,
)
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    default=None,
    help="FQN of the Workspace to deploy to. If not provided, the Workspace FQN will be read from the deployment spec if available.",
)
@click.option(
    "--wait/--no-wait",
    # The leading space is intentional. See: https://github.com/pallets/click/blob/2d610e36a429bfebf0adb0ca90cdc0585f296369/docs/options.md?plain=1#L295
    " /--no_wait",
    is_flag=True,
    show_default=True,
    default=True,
    help="Wait and tail the deployment progress",
)
@click.option(
    "--force/--no-force",
    is_flag=True,
    show_default=True,
    default=False,
    help="Force create a new deployment by canceling any ongoing deployments",
)
@click.option(
    "--trigger-on-deploy/--no-trigger-on-deploy",
    "--trigger_on_deploy/--no_trigger_on_deploy",
    is_flag=True,
    show_default=True,
    default=False,
    help="Trigger a Job run after deployment succeeds. Has no effect for non Job type deployments",
)
@click.pass_context
@handle_exception_wrapper
def deploy_command(
    ctx: click.Context,
    file: str,
    workspace_fqn: Optional[str],
    wait: bool,
    force: bool = False,
    trigger_on_deploy: bool = False,
):
    if ctx.invoked_subcommand is not None:
        return
    from truefoundry.common.session import Session
    from truefoundry.deploy.v2.lib.deployable_patched_models import Application

    try:
        _ = Session.new()
    except Exception as e:
        raise ClickException(message=str(e)) from e

    if file:
        with open(file, "r") as f:
            application_definition = yaml.safe_load(f)

        application = Application.parse_obj(application_definition)
        application.deploy(
            workspace_fqn=workspace_fqn,
            wait=wait,
            force=force,
            trigger_on_deploy=trigger_on_deploy,
        )
        sys.exit(0)

    click.echo(
        click.style(
            "We did not find any truefoundry.yaml or servicefoundry.yaml at the root path.",
            fg="red",
        ),
        color=True,
    )

    if not sys.stdout.isatty():
        click.echo(
            click.style(
                'Please create a truefoundry.yaml or pass the file name with "--file file_name"',
                fg="yellow",
            ),
            color=True,
        )
        sys.exit(1)

    try:
        autodeploy_cli(project_root_path=".", deploy=True, workspace_fqn=workspace_fqn)
    except InvalidRequirementsException as e:
        raise UsageError(message=e.message) from e
    except Exception as e:
        raise UsageError(message=str(e)) from e


@deploy_command.command(name="workflow", cls=COMMAND_CLS, help="Deploy a Workflow")
@click.option(
    "-n",
    "--name",
    required=True,
    help="Name of the Workflow",
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
    help="Path to the Workflow file. e.g. workflow.py",
)
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    required=True,
    help="FQN of the Workspace to deploy to.",
)
@handle_exception_wrapper
def deploy_workflow_command(name: str, file: str, workspace_fqn: str):
    from truefoundry.common.session import Session

    try:
        _ = Session.new()
    except Exception as e:
        raise ClickException(message=str(e)) from e

    from truefoundry.deploy.v2.lib.deployable_patched_models import Workflow

    workflow = Workflow(
        name=name,
        workflow_file_path=file,
    )
    workflow.deploy(workspace_fqn=workspace_fqn)


def get_deploy_command():
    return deploy_command



from dataclasses import dataclass, is_dataclass
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import numpy as np

if TYPE_CHECKING:
    import pandas as pd


from .schema import ParamSchema, Schema, convert_dataclass_to_schema
from .utils import infer_param_schema, infer_schema

MlflowInferableDataset = Union["pd.DataFrame", "np.ndarray", Dict[str, "np.ndarray"]]


class ModelSignature:
    """
    ModelSignature specifies schema of model's inputs, outputs and params.

    ModelSignature can be :py:func:`inferred <mlflow.models.infer_signature>` from training
    dataset, model predictions using and params for inference, or constructed by hand by
    passing an input and output :py:class:`Schema <mlflow.types.Schema>`, and params
    :py:class:`ParamSchema <mlflow.types.ParamSchema>`.
    """

    def __init__(
        self,
        inputs: Union[Schema, dataclass] = None,
        outputs: Union[Schema, dataclass] = None,
        params: ParamSchema = None,
    ):
        if inputs and not isinstance(inputs, Schema) and not is_dataclass(inputs):
            raise TypeError(
                "inputs must be either None, mlflow.models.signature.Schema, or a dataclass,"
                f"got '{type(inputs).__name__}'"
            )
        if outputs and not isinstance(outputs, Schema) and not is_dataclass(outputs):
            raise TypeError(
                "outputs must be either None, mlflow.models.signature.Schema, or a dataclass,"
                f"got '{type(outputs).__name__}'"
            )
        if params and not isinstance(params, ParamSchema):
            raise TypeError(
                "If params are provided, they must by of type mlflow.models.signature.ParamSchema, "
                f"got '{type(params).__name__}'"
            )
        if all(x is None for x in [inputs, outputs, params]):
            raise ValueError(
                "At least one of inputs, outputs or params must be provided"
            )
        if is_dataclass(inputs):
            self.inputs = convert_dataclass_to_schema(inputs)
        else:
            self.inputs = inputs
        if is_dataclass(outputs):
            self.outputs = convert_dataclass_to_schema(outputs)
        else:
            self.outputs = outputs
        self.params = params

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize into a 'jsonable' dictionary.

        Input and output schema are represented as json strings. This is so that the
        representation is compact when embedded in an MLmodel yaml file.

        Returns:
            dictionary representation with input and output schema represented as json strings.
        """

        return {
            "inputs": self.inputs.to_json() if self.inputs else None,
            "outputs": self.outputs.to_json() if self.outputs else None,
            "params": self.params.to_json() if self.params else None,
        }

    @classmethod
    def from_dict(cls, signature_dict: Dict[str, Any]):
        """
        Deserialize from dictionary representation.

        Args:
            signature_dict: Dictionary representation of model signature.
                Expected dictionary format:
                `{'inputs': <json string>,
                'outputs': <json string>,
                'params': <json string>" }`

        Returns:
            ModelSignature populated with the data form the dictionary.
        """
        inputs = Schema.from_json(x) if (x := signature_dict.get("inputs")) else None
        outputs = Schema.from_json(x) if (x := signature_dict.get("outputs")) else None
        params = (
            ParamSchema.from_json(x) if (x := signature_dict.get("params")) else None
        )

        return cls(inputs, outputs, params)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, ModelSignature)
            and self.inputs == other.inputs
            and self.outputs == other.outputs
            and self.params == other.params
        )

    def __repr__(self) -> str:
        return (
            "inputs: \n"
            f"  {self.inputs!r}\n"
            "outputs: \n"
            f"  {self.outputs!r}\n"
            "params: \n"
            f"  {self.params!r}\n"
        )


def infer_signature(
    model_input: Any = None,
    model_output: "MlflowInferableDataset" = None,
    params: Optional[Dict[str, Any]] = None,
) -> ModelSignature:
    if model_input is not None:
        inputs = (
            convert_dataclass_to_schema(model_input)
            if is_dataclass(model_input)
            else infer_schema(model_input)
        )
    else:
        inputs = None
    if model_output is not None:
        outputs = (
            convert_dataclass_to_schema(model_output)
            if is_dataclass(model_output)
            else infer_schema(model_output)
        )
    else:
        outputs = None
    params = infer_param_schema(params) if params else None
    return ModelSignature(inputs, outputs, params)



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from truefoundry.pydantic_v1 import BaseModel, Field, StrictStr, validator


class TensorFlowFramework(BaseModel):
    """
    TensorFlowFramework
    """

    type: StrictStr = Field(default=..., description="Type of the framework")
    __properties = ["type"]

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value not in ("tensorflow",):
            raise ValueError("must be one of enum values ('tensorflow')")
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> TensorFlowFramework:
        """Create an instance of TensorFlowFramework from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> TensorFlowFramework:
        """Create an instance of TensorFlowFramework from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return TensorFlowFramework.parse_obj(obj)

        _obj = TensorFlowFramework.parse_obj({"type": obj.get("type")})
        return _obj



from typing import List, Optional, Tuple

import rich_click as click
import yaml

from truefoundry.cli.console import console
from truefoundry.cli.const import GROUP_CLS
from truefoundry.cli.util import handle_exception_wrapper
from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)
from truefoundry.deploy.lib.dao import apply as apply_lib
from truefoundry.deploy.lib.dao import delete as delete_lib
from truefoundry.deploy.lib.model.entity import (
    File,
    ResolvedDirectoryFiles,
    ResolvedResourceInfo,
)


def _validate_options(
    files: Tuple[str, ...],
    directory: Optional[str],
    diffs_only: bool,
    ref: Optional[str],
    show_diff: bool,
    dry_run: bool,
    sync: bool,
) -> None:
    if show_diff and not dry_run:
        raise click.ClickException("--show-diff requires --dry-run")

    if ref is not None and not diffs_only:
        raise click.ClickException("--ref requires --diffs-only")

    if diffs_only and not directory:
        raise click.ClickException("--diffs-only requires -d/--directory")

    if files and directory:
        raise click.ClickException(
            "-f/--file and -d/--directory are mutually exclusive"
        )

    if not files and not directory:
        raise click.ClickException("Either -f/--file or -d/--directory is required")

    if sync and (not directory or not diffs_only):
        raise click.ClickException("--sync requires -d/--directory and --diffs-only")


def _resolve_directory_files(
    directory: str, diffs_only: bool, sync: bool, dry_run: bool, ref: Optional[str]
) -> Optional[ResolvedDirectoryFiles]:
    deleted_files: List[File] = []
    if diffs_only:
        git_changed_yaml_files = apply_lib.get_git_changed_yaml_files(directory, ref)
        resolved_files = git_changed_yaml_files.modified
        deleted_files = git_changed_yaml_files.deleted

        ref_label = ref or "HEAD"
        if not resolved_files and not deleted_files:
            console.print(
                f"[yellow]No changed yaml files found compared to {ref_label}[/]"
            )
            return None
        console.print(
            f"[cyan]Found {len(resolved_files)} changed yaml file(s) and {len(deleted_files)} deleted yaml file(s) compared to {ref_label}[/]"
        )
    else:
        resolved_file_paths = apply_lib.collect_yaml_files(directory)
        resolved_files = [File(path=file_path) for file_path in resolved_file_paths]
        if not resolved_files:
            console.print("[yellow]No yaml files found in the directory[/]")
            return None
        console.print(f"[cyan]Found {len(resolved_files)} yaml file(s) in directory[/]")

    if sync and deleted_files:
        suffix = ""
        if dry_run:
            suffix = "No changes will be made right now."
        else:
            suffix = "Resources from these yaml files will be deleted."
        console.print(f"[cyan]The following files were deleted. {suffix}[/]")
        for file in deleted_files:
            try:
                manifest = yaml.safe_load(file.content)
            except Exception:
                continue
            manifest_type_and_name = apply_lib.get_type_and_name_from_manifest(manifest)
            if manifest_type_and_name is not None:
                console.print(
                    f"  [dim]{file.path}[/] [cyan](type: {manifest_type_and_name.type}, name: {manifest_type_and_name.name})[/]"
                )
            else:
                console.print(
                    f"  [dim]{file.path}[/] [cyan](type: unknown, name: unknown)[/]"
                )
        console.print()

    return ResolvedDirectoryFiles(
        modified=resolved_files,
        deleted=deleted_files,
    )


def _log_dependency_order(sorted_resolved_files: List[str], dry_run: bool) -> None:
    if not dry_run:
        return

    if len(sorted_resolved_files) == 0:
        return

    console.print("[cyan]Applying in dependency order:[/]")
    for f in sorted_resolved_files:
        manifest_type_and_name = apply_lib.read_manifest_type_and_name(f)
        type_str = (
            manifest_type_and_name.type
            if manifest_type_and_name is not None
            else "unknown"
        )
        console.print(f"  [dim]{f}[/] [cyan]({type_str})[/]")
    console.print()


def _log_skipped_resources(
    skipped_resources: List[ResolvedResourceInfo], dry_run: bool
) -> None:
    if not dry_run or not skipped_resources:
        return

    console.print(
        f"[yellow]Skipped dry run for {len(skipped_resources)} resource(s) that have dependencies on other resources:[/]"
    )
    for resource in skipped_resources:
        console.print(
            f"  [dim]Type: {resource.type}, Name: {resource.name if resource.name else '(no name specified)'}[/]",
            highlight=False,
        )


@click.group(
    name="apply",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="""Create resources by applying manifest locally from TrueFoundry spec.

\b
Examples:
  tfy apply -f manifest.yaml                        Apply a single manifest file
  tfy apply -f team.yaml -f cluster.yaml            Apply multiple files
  tfy apply -d ./manifests                          Apply all yamls in a directory (sorted by dependency order)
  tfy apply -d ./manifests --diffs-only             Apply only git-changed yamls compared to HEAD
  tfy apply -d ./manifests --diffs-only --ref main  Apply only git-changed yamls compared to a branch/tag/commit
  tfy apply -d ./manifests --dry-run                Dry run (simulate without applying)
  tfy apply -d ./manifests --dry-run --show-diff    Dry run with diff output
  tfy apply -d ./manifests --diffs-only --sync      Sync resources by deleting those not present in the manifests
""",
    context_settings={"ignore_unknown_options": True, "allow_interspersed_args": True},
)
@click.option(
    "-f",
    "--file",
    "files",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to yaml manifest file (You can apply multiple files at once by providing multiple -f options)",
    show_default=True,
    multiple=True,
)
@click.option(
    "-d",
    "--directory",
    "directory",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Path to a directory containing yaml manifest files. All yamls will be collected and applied in dependency order.",
)
@click.option(
    "--diffs-only",
    "--diffs_only",
    is_flag=True,
    show_default=True,
    help="Only apply yaml files that have changed in git (requires -d)",
)
@click.option(
    "--ref",
    type=str,
    default=None,
    help="Git ref (branch, tag, or commit) to compare against when using --diffs-only (default: HEAD)",
)
@click.option(
    "--dry-run",
    "--dry_run",
    is_flag=True,
    show_default=True,
    help="Simulate the process without actually applying the manifest",
)
@click.option(
    "--show-diff",
    "--show_diff",
    is_flag=True,
    show_default=True,
    help="Print manifest differences when using --dry-run",
)
@click.option(
    "--sync",
    is_flag=True,
    show_default=True,
    help="Sync resources by deleting those not present in the manifests (requires -d/--directory and --diffs-only)",
)
@handle_exception_wrapper
def apply_command(
    files: Tuple[str, ...] = (),
    directory: Optional[str] = None,
    diffs_only: bool = False,
    ref: Optional[str] = None,
    dry_run: bool = False,
    show_diff: bool = False,
    sync: bool = False,
):
    _validate_options(files, directory, diffs_only, ref, show_diff, dry_run, sync)

    resolved_file_paths: Optional[List[str]] = None

    if directory:
        resolved_directory_files = _resolve_directory_files(
            directory, diffs_only, sync, dry_run, ref
        )
        if resolved_directory_files is None:
            return
        resolved_file_paths = [file.path for file in resolved_directory_files.modified]
        deleted_files = resolved_directory_files.deleted
    else:
        resolved_file_paths = list(files)

    sf_server_client = ServiceFoundryServiceClient()

    skipped_resources: List[ResolvedResourceInfo] = []
    if dry_run:
        resolved_resources = apply_lib.resolve_resources_for_dry_run(
            resolved_file_paths, sf_server_client
        )
        resolved_file_paths = [
            resource.filepath for resource in resolved_resources.processable_resources
        ]
        skipped_resources = resolved_resources.skipped_resources

    # Sort the resolved files by manifest type according to dependency order
    # TODO: Avoid reading from file again here; save the manifest type and name beforehand with resolved_file_paths
    resolved_file_paths = apply_lib.sort_files_by_manifest_type(resolved_file_paths)
    _log_dependency_order(resolved_file_paths, dry_run)

    error_message = None

    apply_results = apply_lib.apply_manifest_files(
        resolved_file_paths, sf_server_client, dry_run, show_diff
    )

    _log_skipped_resources(skipped_resources, dry_run)

    if not all(apply_result.success for apply_result in apply_results):
        error_message = "Failed to apply one or more resource manifests"

    if sync and deleted_files and not dry_run:
        delete_results = delete_lib.delete_manifest_files(
            files=deleted_files,
            sf_server_client=sf_server_client,
            read_content_from_file=False,
        )

        if not all(delete_result.success for delete_result in delete_results):
            if error_message is None:
                error_message = "Failed to delete one or more resource manifests"
            else:
                error_message = (
                    "Failed to apply and delete one or more resource manifests"
                )

    if error_message:
        raise Exception(error_message)


def get_apply_command():
    return apply_command



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.ml._autogen.client.models.run_data_dto import RunDataDto
from truefoundry.ml._autogen.client.models.run_info_dto import RunInfoDto
from truefoundry.pydantic_v1 import BaseModel, Field


class RunDto(BaseModel):
    """
    RunDto
    """

    info: RunInfoDto = Field(...)
    data: Optional[RunDataDto] = None
    __properties = ["info", "data"]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> RunDto:
        """Create an instance of RunDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of info
        if self.info:
            _dict["info"] = self.info.to_dict()
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of data
        if self.data:
            _dict["data"] = self.data.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> RunDto:
        """Create an instance of RunDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return RunDto.parse_obj(obj)

        _obj = RunDto.parse_obj(
            {
                "info": RunInfoDto.from_dict(obj.get("info"))
                if obj.get("info") is not None
                else None,
                "data": RunDataDto.from_dict(obj.get("data"))
                if obj.get("data") is not None
                else None,
            }
        )
        return _obj



# coding: utf-8

# flake8: noqa
"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

# import models into model package
from truefoundry.ml._autogen.client.models.agent_app import AgentApp
from truefoundry.ml._autogen.client.models.agent_manifest import AgentManifest
from truefoundry.ml._autogen.client.models.agent_open_api_tool_manifest import (
    AgentOpenAPIToolManifest,
)
from truefoundry.ml._autogen.client.models.agent_open_api_tool_with_fqn import (
    AgentOpenAPIToolWithFQN,
)
from truefoundry.ml._autogen.client.models.agent_with_fqn import AgentWithFQN
from truefoundry.ml._autogen.client.models.apply_request_dto import ApplyRequestDto
from truefoundry.ml._autogen.client.models.apply_response_dto import ApplyResponseDto
from truefoundry.ml._autogen.client.models.artifact_dto import ArtifactDto
from truefoundry.ml._autogen.client.models.artifact_manifest import ArtifactManifest
from truefoundry.ml._autogen.client.models.artifact_path import ArtifactPath
from truefoundry.ml._autogen.client.models.artifact_response_dto import (
    ArtifactResponseDto,
)
from truefoundry.ml._autogen.client.models.artifact_type import ArtifactType
from truefoundry.ml._autogen.client.models.artifact_version_dto import (
    ArtifactVersionDto,
)
from truefoundry.ml._autogen.client.models.artifact_version_response_dto import (
    ArtifactVersionResponseDto,
)
from truefoundry.ml._autogen.client.models.artifact_version_serialization_format import (
    ArtifactVersionSerializationFormat,
)
from truefoundry.ml._autogen.client.models.artifact_version_status import (
    ArtifactVersionStatus,
)
from truefoundry.ml._autogen.client.models.assistant_message import AssistantMessage
from truefoundry.ml._autogen.client.models.authorize_user_for_model_request_dto import (
    AuthorizeUserForModelRequestDto,
)
from truefoundry.ml._autogen.client.models.authorize_user_for_model_version_request_dto import (
    AuthorizeUserForModelVersionRequestDto,
)
from truefoundry.ml._autogen.client.models.blob_storage_reference import (
    BlobStorageReference,
)
from truefoundry.ml._autogen.client.models.body_get_search_runs_get import (
    BodyGetSearchRunsGet,
)
from truefoundry.ml._autogen.client.models.chat_prompt_manifest import (
    ChatPromptManifest,
)
from truefoundry.ml._autogen.client.models.chat_prompt_manifest_messages_inner import (
    ChatPromptManifestMessagesInner,
)
from truefoundry.ml._autogen.client.models.columns_dto import ColumnsDto
from truefoundry.ml._autogen.client.models.command import Command
from truefoundry.ml._autogen.client.models.content import Content
from truefoundry.ml._autogen.client.models.content1 import Content1
from truefoundry.ml._autogen.client.models.content2 import Content2
from truefoundry.ml._autogen.client.models.content2_any_of_inner import (
    Content2AnyOfInner,
)
from truefoundry.ml._autogen.client.models.create_artifact_request_dto import (
    CreateArtifactRequestDto,
)
from truefoundry.ml._autogen.client.models.create_artifact_response_dto import (
    CreateArtifactResponseDto,
)
from truefoundry.ml._autogen.client.models.create_artifact_version_request_dto import (
    CreateArtifactVersionRequestDto,
)
from truefoundry.ml._autogen.client.models.create_artifact_version_response_dto import (
    CreateArtifactVersionResponseDto,
)
from truefoundry.ml._autogen.client.models.create_dataset_request_dto import (
    CreateDatasetRequestDto,
)
from truefoundry.ml._autogen.client.models.create_experiment_request_dto import (
    CreateExperimentRequestDto,
)
from truefoundry.ml._autogen.client.models.create_experiment_response_dto import (
    CreateExperimentResponseDto,
)
from truefoundry.ml._autogen.client.models.create_model_version_request_dto import (
    CreateModelVersionRequestDto,
)
from truefoundry.ml._autogen.client.models.create_multi_part_upload_for_dataset_request_dto import (
    CreateMultiPartUploadForDatasetRequestDto,
)
from truefoundry.ml._autogen.client.models.create_multi_part_upload_for_dataset_response_dto import (
    CreateMultiPartUploadForDatasetResponseDto,
)
from truefoundry.ml._autogen.client.models.create_multi_part_upload_request_dto import (
    CreateMultiPartUploadRequestDto,
)
from truefoundry.ml._autogen.client.models.create_python_deployment_config_request_dto import (
    CreatePythonDeploymentConfigRequestDto,
)
from truefoundry.ml._autogen.client.models.create_python_deployment_config_response_dto import (
    CreatePythonDeploymentConfigResponseDto,
)
from truefoundry.ml._autogen.client.models.create_run_request_dto import (
    CreateRunRequestDto,
)
from truefoundry.ml._autogen.client.models.create_run_response_dto import (
    CreateRunResponseDto,
)
from truefoundry.ml._autogen.client.models.create_workflow_task_config_request_dto import (
    CreateWorkflowTaskConfigRequestDto,
)
from truefoundry.ml._autogen.client.models.data_directory_manifest import (
    DataDirectoryManifest,
)
from truefoundry.ml._autogen.client.models.dataset_dto import DatasetDto
from truefoundry.ml._autogen.client.models.dataset_response_dto import (
    DatasetResponseDto,
)
from truefoundry.ml._autogen.client.models.delete_artifact_versions_request_dto import (
    DeleteArtifactVersionsRequestDto,
)
from truefoundry.ml._autogen.client.models.delete_dataset_request_dto import (
    DeleteDatasetRequestDto,
)
from truefoundry.ml._autogen.client.models.delete_files_for_dataset_request_dto import (
    DeleteFilesForDatasetRequestDto,
)
from truefoundry.ml._autogen.client.models.delete_model_version_request_dto import (
    DeleteModelVersionRequestDto,
)
from truefoundry.ml._autogen.client.models.delete_run_request import DeleteRunRequest
from truefoundry.ml._autogen.client.models.delete_tag_request_dto import (
    DeleteTagRequestDto,
)
from truefoundry.ml._autogen.client.models.experiment_dto import ExperimentDto
from truefoundry.ml._autogen.client.models.experiment_id_request_dto import (
    ExperimentIdRequestDto,
)
from truefoundry.ml._autogen.client.models.experiment_response_dto import (
    ExperimentResponseDto,
)
from truefoundry.ml._autogen.client.models.experiment_tag_dto import ExperimentTagDto
from truefoundry.ml._autogen.client.models.export_deployment_files_request_dto import (
    ExportDeploymentFilesRequestDto,
)
from truefoundry.ml._autogen.client.models.external_blob_storage_source import (
    ExternalBlobStorageSource,
)
from truefoundry.ml._autogen.client.models.fast_ai_framework import FastAIFramework
from truefoundry.ml._autogen.client.models.file_info_dto import FileInfoDto
from truefoundry.ml._autogen.client.models.finalize_artifact_version_request_dto import (
    FinalizeArtifactVersionRequestDto,
)
from truefoundry.ml._autogen.client.models.framework import Framework
from truefoundry.ml._autogen.client.models.get_artifact_version_aliases_response_dto import (
    GetArtifactVersionAliasesResponseDto,
)
from truefoundry.ml._autogen.client.models.get_experiment_response_dto import (
    GetExperimentResponseDto,
)
from truefoundry.ml._autogen.client.models.get_latest_run_log_response_dto import (
    GetLatestRunLogResponseDto,
)
from truefoundry.ml._autogen.client.models.get_metric_history_response import (
    GetMetricHistoryResponse,
)
from truefoundry.ml._autogen.client.models.get_signed_url_for_dataset_write_request_dto import (
    GetSignedURLForDatasetWriteRequestDto,
)
from truefoundry.ml._autogen.client.models.get_signed_urls_for_artifact_version_read_request_dto import (
    GetSignedURLsForArtifactVersionReadRequestDto,
)
from truefoundry.ml._autogen.client.models.get_signed_urls_for_artifact_version_read_response_dto import (
    GetSignedURLsForArtifactVersionReadResponseDto,
)
from truefoundry.ml._autogen.client.models.get_signed_urls_for_artifact_version_write_request_dto import (
    GetSignedURLsForArtifactVersionWriteRequestDto,
)
from truefoundry.ml._autogen.client.models.get_signed_urls_for_artifact_version_write_response_dto import (
    GetSignedURLsForArtifactVersionWriteResponseDto,
)
from truefoundry.ml._autogen.client.models.get_signed_urls_for_dataset_read_request_dto import (
    GetSignedURLsForDatasetReadRequestDto,
)
from truefoundry.ml._autogen.client.models.get_signed_urls_for_dataset_read_response_dto import (
    GetSignedURLsForDatasetReadResponseDto,
)
from truefoundry.ml._autogen.client.models.get_signed_urls_for_dataset_write_response_dto import (
    GetSignedURLsForDatasetWriteResponseDto,
)
from truefoundry.ml._autogen.client.models.get_tenant_id_response_dto import (
    GetTenantIdResponseDto,
)
from truefoundry.ml._autogen.client.models.gluon_framework import GluonFramework
from truefoundry.ml._autogen.client.models.h2_o_framework import H2OFramework
from truefoundry.ml._autogen.client.models.http_validation_error import (
    HTTPValidationError,
)
from truefoundry.ml._autogen.client.models.image_content_part import ImageContentPart
from truefoundry.ml._autogen.client.models.image_url import ImageUrl
from truefoundry.ml._autogen.client.models.infer_method_name import InferMethodName
from truefoundry.ml._autogen.client.models.keras_framework import KerasFramework
from truefoundry.ml._autogen.client.models.latest_run_log_dto import LatestRunLogDto
from truefoundry.ml._autogen.client.models.library_name import LibraryName
from truefoundry.ml._autogen.client.models.light_gbm_framework import LightGBMFramework
from truefoundry.ml._autogen.client.models.list_artifact_versions_request_dto import (
    ListArtifactVersionsRequestDto,
)
from truefoundry.ml._autogen.client.models.list_artifact_versions_response_dto import (
    ListArtifactVersionsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_artifacts_request_dto import (
    ListArtifactsRequestDto,
)
from truefoundry.ml._autogen.client.models.list_artifacts_response_dto import (
    ListArtifactsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_colums_response_dto import (
    ListColumsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_datasets_request_dto import (
    ListDatasetsRequestDto,
)
from truefoundry.ml._autogen.client.models.list_datasets_response_dto import (
    ListDatasetsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_experiments_response_dto import (
    ListExperimentsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_files_for_artifact_version_request_dto import (
    ListFilesForArtifactVersionRequestDto,
)
from truefoundry.ml._autogen.client.models.list_files_for_artifact_versions_response_dto import (
    ListFilesForArtifactVersionsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_files_for_dataset_request_dto import (
    ListFilesForDatasetRequestDto,
)
from truefoundry.ml._autogen.client.models.list_files_for_dataset_response_dto import (
    ListFilesForDatasetResponseDto,
)
from truefoundry.ml._autogen.client.models.list_latest_run_logs_response_dto import (
    ListLatestRunLogsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_metric_history_request_dto import (
    ListMetricHistoryRequestDto,
)
from truefoundry.ml._autogen.client.models.list_metric_history_response_dto import (
    ListMetricHistoryResponseDto,
)
from truefoundry.ml._autogen.client.models.list_model_version_response_dto import (
    ListModelVersionResponseDto,
)
from truefoundry.ml._autogen.client.models.list_model_versions_request_dto import (
    ListModelVersionsRequestDto,
)
from truefoundry.ml._autogen.client.models.list_models_request_dto import (
    ListModelsRequestDto,
)
from truefoundry.ml._autogen.client.models.list_models_response_dto import (
    ListModelsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_run_artifacts_response_dto import (
    ListRunArtifactsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_run_logs_response_dto import (
    ListRunLogsResponseDto,
)
from truefoundry.ml._autogen.client.models.local_artifact_source import (
    LocalArtifactSource,
)
from truefoundry.ml._autogen.client.models.local_model_source import LocalModelSource
from truefoundry.ml._autogen.client.models.log_batch_request_dto import (
    LogBatchRequestDto,
)
from truefoundry.ml._autogen.client.models.log_metric_request_dto import (
    LogMetricRequestDto,
)
from truefoundry.ml._autogen.client.models.log_param_request_dto import (
    LogParamRequestDto,
)
from truefoundry.ml._autogen.client.models.manifest import Manifest
from truefoundry.ml._autogen.client.models.manifest1 import Manifest1
from truefoundry.ml._autogen.client.models.manifest2 import Manifest2
from truefoundry.ml._autogen.client.models.method import Method
from truefoundry.ml._autogen.client.models.metric_collection_dto import (
    MetricCollectionDto,
)
from truefoundry.ml._autogen.client.models.metric_dto import MetricDto
from truefoundry.ml._autogen.client.models.mime_type import MimeType
from truefoundry.ml._autogen.client.models.model_configuration import ModelConfiguration
from truefoundry.ml._autogen.client.models.model_dto import ModelDto
from truefoundry.ml._autogen.client.models.model_manifest import ModelManifest
from truefoundry.ml._autogen.client.models.model_response_dto import ModelResponseDto
from truefoundry.ml._autogen.client.models.model_server import ModelServer
from truefoundry.ml._autogen.client.models.model_version_dto import ModelVersionDto
from truefoundry.ml._autogen.client.models.model_version_environment import (
    ModelVersionEnvironment,
)
from truefoundry.ml._autogen.client.models.model_version_response_dto import (
    ModelVersionResponseDto,
)
from truefoundry.ml._autogen.client.models.multi_part_upload_dto import (
    MultiPartUploadDto,
)
from truefoundry.ml._autogen.client.models.multi_part_upload_response_dto import (
    MultiPartUploadResponseDto,
)
from truefoundry.ml._autogen.client.models.multi_part_upload_storage_provider import (
    MultiPartUploadStorageProvider,
)
from truefoundry.ml._autogen.client.models.notify_artifact_version_failure_dto import (
    NotifyArtifactVersionFailureDto,
)
from truefoundry.ml._autogen.client.models.onnx_framework import ONNXFramework
from truefoundry.ml._autogen.client.models.openapi_spec import OpenapiSpec
from truefoundry.ml._autogen.client.models.paddle_framework import PaddleFramework
from truefoundry.ml._autogen.client.models.param_dto import ParamDto
from truefoundry.ml._autogen.client.models.parameters import Parameters
from truefoundry.ml._autogen.client.models.py_torch_framework import PyTorchFramework
from truefoundry.ml._autogen.client.models.resolve_agent_app_response_dto import (
    ResolveAgentAppResponseDto,
)
from truefoundry.ml._autogen.client.models.restore_run_request_dto import (
    RestoreRunRequestDto,
)
from truefoundry.ml._autogen.client.models.run_data_dto import RunDataDto
from truefoundry.ml._autogen.client.models.run_dto import RunDto
from truefoundry.ml._autogen.client.models.run_info_dto import RunInfoDto
from truefoundry.ml._autogen.client.models.run_log_dto import RunLogDto
from truefoundry.ml._autogen.client.models.run_log_input_dto import RunLogInputDto
from truefoundry.ml._autogen.client.models.run_response_dto import RunResponseDto
from truefoundry.ml._autogen.client.models.run_tag_dto import RunTagDto
from truefoundry.ml._autogen.client.models.search_runs_request_dto import (
    SearchRunsRequestDto,
)
from truefoundry.ml._autogen.client.models.search_runs_response_dto import (
    SearchRunsResponseDto,
)
from truefoundry.ml._autogen.client.models.set_experiment_tag_request_dto import (
    SetExperimentTagRequestDto,
)
from truefoundry.ml._autogen.client.models.set_tag_request_dto import SetTagRequestDto
from truefoundry.ml._autogen.client.models.signed_url_dto import SignedURLDto
from truefoundry.ml._autogen.client.models.sklearn_framework import SklearnFramework
from truefoundry.ml._autogen.client.models.sklearn_model_schema import (
    SklearnModelSchema,
)
from truefoundry.ml._autogen.client.models.sklearn_serialization_format import (
    SklearnSerializationFormat,
)
from truefoundry.ml._autogen.client.models.source import Source
from truefoundry.ml._autogen.client.models.source1 import Source1
from truefoundry.ml._autogen.client.models.source2 import Source2
from truefoundry.ml._autogen.client.models.spa_cy_framework import SpaCyFramework
from truefoundry.ml._autogen.client.models.stats_models_framework import (
    StatsModelsFramework,
)
from truefoundry.ml._autogen.client.models.stop import Stop
from truefoundry.ml._autogen.client.models.store_run_logs_request_dto import (
    StoreRunLogsRequestDto,
)
from truefoundry.ml._autogen.client.models.subject import Subject
from truefoundry.ml._autogen.client.models.subject_type import SubjectType
from truefoundry.ml._autogen.client.models.system_message import SystemMessage
from truefoundry.ml._autogen.client.models.tensor_flow_framework import (
    TensorFlowFramework,
)
from truefoundry.ml._autogen.client.models.text import Text
from truefoundry.ml._autogen.client.models.text_content_part import TextContentPart
from truefoundry.ml._autogen.client.models.transformers_framework import (
    TransformersFramework,
)
from truefoundry.ml._autogen.client.models.trigger_job_run_config_request_dto import (
    TriggerJobRunConfigRequestDto,
)
from truefoundry.ml._autogen.client.models.trigger_job_run_config_response_dto import (
    TriggerJobRunConfigResponseDto,
)
from truefoundry.ml._autogen.client.models.true_foundry_managed_source import (
    TrueFoundryManagedSource,
)
from truefoundry.ml._autogen.client.models.update_artifact_version_request_dto import (
    UpdateArtifactVersionRequestDto,
)
from truefoundry.ml._autogen.client.models.update_dataset_request_dto import (
    UpdateDatasetRequestDto,
)
from truefoundry.ml._autogen.client.models.update_experiment_request_dto import (
    UpdateExperimentRequestDto,
)
from truefoundry.ml._autogen.client.models.update_model_version_request_dto import (
    UpdateModelVersionRequestDto,
)
from truefoundry.ml._autogen.client.models.update_run_request_dto import (
    UpdateRunRequestDto,
)
from truefoundry.ml._autogen.client.models.update_run_response_dto import (
    UpdateRunResponseDto,
)
from truefoundry.ml._autogen.client.models.url import Url
from truefoundry.ml._autogen.client.models.user_message import UserMessage
from truefoundry.ml._autogen.client.models.validate_external_storage_root_request_dto import (
    ValidateExternalStorageRootRequestDto,
)
from truefoundry.ml._autogen.client.models.validate_external_storage_root_response_dto import (
    ValidateExternalStorageRootResponseDto,
)
from truefoundry.ml._autogen.client.models.validation_error import ValidationError
from truefoundry.ml._autogen.client.models.validation_error_loc_inner import (
    ValidationErrorLocInner,
)
from truefoundry.ml._autogen.client.models.xg_boost_framework import XGBoostFramework
from truefoundry.ml._autogen.client.models.xg_boost_model_schema import (
    XGBoostModelSchema,
)
from truefoundry.ml._autogen.client.models.xg_boost_serialization_format import (
    XGBoostSerializationFormat,
)



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, Dict, Optional

from truefoundry.ml._autogen.client.models.source1 import Source1
from truefoundry.pydantic_v1 import BaseModel, Field, StrictStr, constr, validator


class DataDirectoryManifest(BaseModel):
    """
    Data Directory manifest.  # noqa: E501
    """

    type: Optional[StrictStr] = "data-dir"
    name: constr(strict=True, max_length=256) = Field(
        default=..., description="Name of the data directory"
    )
    ml_repo: constr(strict=True) = Field(
        default=...,
        description="Name of the ML Repo to create the data directory under",
    )
    description: Optional[constr(strict=True, max_length=512)] = Field(
        default=None, description="Description for the data directory"
    )
    metadata: Dict[str, Any] = Field(
        default=...,
        description='Key Value metadata. Should be valid JSON. For e.g. `{"business-unit": "sales", "quality": "good", "rating": 4.5}`',
    )
    source: Source1 = Field(...)
    __properties = ["type", "name", "ml_repo", "description", "metadata", "source"]

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in ("data-dir",):
            raise ValueError("must be one of enum values ('data-dir')")
        return value

    @validator("name")
    def name_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^[A-Za-z0-9_\-]+$", value):
            raise ValueError(
                r"must validate the regular expression /^[A-Za-z0-9_\-]+$/"
            )
        return value

    @validator("ml_repo")
    def ml_repo_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9\-]{1,98}[a-zA-Z0-9]$", value):
            raise ValueError(
                r"must validate the regular expression /^[a-zA-Z][a-zA-Z0-9\-]{1,98}[a-zA-Z0-9]$/"
            )
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> DataDirectoryManifest:
        """Create an instance of DataDirectoryManifest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of source
        if self.source:
            _dict["source"] = self.source.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> DataDirectoryManifest:
        """Create an instance of DataDirectoryManifest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return DataDirectoryManifest.parse_obj(obj)

        _obj = DataDirectoryManifest.parse_obj(
            {
                "type": obj.get("type") if obj.get("type") is not None else "data-dir",
                "name": obj.get("name"),
                "ml_repo": obj.get("ml_repo"),
                "description": obj.get("description"),
                "metadata": obj.get("metadata"),
                "source": Source1.from_dict(obj.get("source"))
                if obj.get("source") is not None
                else None,
            }
        )
        return _obj



import collections
import copy
import datetime
import tempfile
from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Tuple, Union
from warnings import warn

from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    DatasetDto,
    DeleteDatasetRequestDto,
    MlfoundryArtifactsApi,
    UpdateDatasetRequestDto,
)
from truefoundry.ml.artifact.truefoundry_artifact_repo import (
    ArtifactIdentifier,
    MlFoundryArtifactsRepository,
)
from truefoundry.ml.entities import FileInfo
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.log_types.artifacts.utils import (
    _copy_additional_files,
    _get_src_dest_pairs,
    _validate_artifact_metadata,
    _validate_description,
    calculate_total_size,
)
from truefoundry.ml.logger import logger
from truefoundry.ml.session import _get_api_client


class DataDirectoryPath(NamedTuple):
    src: str
    dest: Optional[str] = None


class DataDirectory:
    def __init__(self, dataset: DatasetDto) -> None:
        self._api_client = _get_api_client()
        self._mlfoundry_artifacts_api = MlfoundryArtifactsApi(
            api_client=self._api_client
        )
        self._dataset = dataset
        self._description: str = ""
        self._deleted = False
        self._metadata: Dict[str, Any] = {}
        self._set_mutable_attrs()

    @classmethod
    def from_fqn(cls, fqn: str):
        """
        Get the DataDirectory to download contents or load them in memory

        Args:
            fqn (str): Fully qualified name of the data directory.

        Returns:
            DataDirectory: An DataDirectory instance of the artifact

        Examples:

            ```python
            from truefoundry.ml import get_client, DataDirectory

            client = get_client()
            data_directory = DataDirectory.from_fqn(fqn="<data_directory-fqn>")
            ```
        """
        api_client = _get_api_client()
        mlfoundry_artifacts_api = MlfoundryArtifactsApi(api_client=api_client)
        _dataset = mlfoundry_artifacts_api.get_dataset_by_fqn_get(fqn=fqn)
        return cls(_dataset.dataset)

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r})"

    def _set_mutable_attrs(self, refetch=False):
        if refetch:
            _dataset = self._mlfoundry_artifacts_api.get_dataset_by_fqn_get(
                fqn=self._dataset.fqn
            )
            self._dataset = _dataset.dataset
        self._description = self._dataset.description or ""
        self._metadata = copy.deepcopy(self._dataset.dataset_metadata)

    def _get_artifacts_repo(self):
        return MlFoundryArtifactsRepository(
            artifact_identifier=ArtifactIdentifier(
                dataset_id=self._dataset.id, dataset_fqn=self._dataset.fqn
            ),
            api_client=self._api_client,
        )

    @property
    def name(self) -> str:
        """Get the name of the DataDirectory"""
        return self._dataset.name

    @property
    def fqn(self) -> str:
        """Get fqn of the DataDirectory"""
        return self._dataset.fqn

    @property
    def storage_root(self) -> str:
        """Get storage_root of the DataDirectory"""
        return self._dataset.storage_root

    @property
    def description(self) -> Optional[str]:
        """Get description of the DataDirectory"""
        return self._description

    @description.setter
    def description(self, value: str):
        """set the description of the DataDirectory"""
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        """Get metadata for the current DataDirectory"""
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        """set the metadata for current DataDirectory"""
        _validate_artifact_metadata(value)
        self._metadata = value

    @property
    def created_at(self) -> datetime.datetime:
        """Get the time at which DataDirectory was created"""
        return self._dataset.created_at

    @property
    def updated_at(self) -> datetime.datetime:
        """Get the information about when the DataDirectory was updated"""
        return self._dataset.updated_at

    def add_files(
        self,
        file_paths: List[
            Union[Tuple[str], Tuple[str, Optional[str]], DataDirectoryPath]
        ],
        progress: Optional[bool] = None,
    ) -> None:
        """Logs File in the `DataDirectory`.

        Args:
            file_paths (List[truefoundry.ml.DataDirectoryPath], optional): A list of pairs
                of (source path, destination path) to add files and folders to the DataDirectory contents.
                The first member of the pair should be a file or directory path and the second member should be
                the path inside the artifact contents to upload to.

                ```python
                from truefoundry.ml import DataDirectoryPath

                data_directory.add_files(
                    file_paths=[
                        DataDirectoryPath("foo.txt", "foo/bar/foo.txt"),
                        DataDirectoryPath("tokenizer/", "foo/tokenizer/"),
                        DataDirectoryPath('bar.text'),
                        ('bar.txt', ),
                        ('foo.txt', 'a/foo.txt')
                    ]
                )
                ```

                would result in

                ```
                .
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
                ```

        Examples:
            ```python
            import os
            from truefoundry.ml import get_client, DataDirectoryPath

            with open("artifact.txt", "w") as f:
                f.write("hello-world")

            client = get_client()
            data_directory = client.get_data_directory_by_fqn(fqn="<data_directory-fqn>")

            data_directory.add_files(
                file_paths=[DataDirectoryPath('artifact.txt', 'a/b/')]
            )
            ```
        """
        for i, file_path in enumerate(file_paths):
            if isinstance(file_path, DataDirectoryPath):
                continue
            elif isinstance(file_path, collections.abc.Sequence) and (
                0 < len(file_path) <= 2
            ):
                file_paths[i] = DataDirectoryPath(*file_path)
            else:
                raise ValueError(
                    "`file_paths` should be an instance of `truefoundry.ml.DataDirectoryPath` "
                    "or a tuple of (src, dest) path strings"
                )

        logger.info("Adding the files to data_directory, this might take a while ...")
        temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

        try:
            logger.info("Copying the files to add")
            temp_dest_to_src_map = _copy_additional_files(
                root_dir=temp_dir.name,
                files_dir="",
                model_dir=None,
                additional_files=file_paths,
            )

        except Exception as e:
            temp_dir.cleanup()
            raise MlFoundryException("Failed to Add Files to DataDirectory") from e

        artifacts_repo = self._get_artifacts_repo()
        total_size = calculate_total_size(list(temp_dest_to_src_map.values()))
        try:
            logger.info(
                "Packaging and uploading files to remote with size: %.6f MB",
                total_size / 1000000.0,
            )
            src_dest_pairs = _get_src_dest_pairs(
                root_dir=temp_dir.name, dest_to_src_map=temp_dest_to_src_map
            )
            artifacts_repo.log_artifacts(
                src_dest_pairs=src_dest_pairs,
                progress=progress,
            )
        except Exception as e:
            raise MlFoundryException("Failed to Add Files to DataDirectory") from e
        finally:
            temp_dir.cleanup()

    def list_files(
        self,
        path: Optional[str] = None,
    ) -> Iterator[FileInfo]:
        """
        List all the files and folders in the data_directory.

        Args:
            path: The path of directory in data_directory, from which the files are to be listed.

        Returns:
            Iterator[FileInfoDto]: List of FileInfo objects that contain path and file_size attributes.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            data_directory = client.get_data_directory_by_fqn(fqn="<your-artifact-fqn>")
            files = data_directory.list_files()
            for file in files:
                print(file.path)
            ```
        """
        artifacts_repo = self._get_artifacts_repo()
        for file_info_dto in artifacts_repo.list_artifacts(path=path):
            yield FileInfo.from_dto(file_info_dto)

    def download(
        self,
        remote_path: Optional[str] = "",
        path: Optional[str] = None,
        overwrite: bool = False,
        progress: Optional[bool] = None,
        download_path: Optional[str] = "",
    ):
        """
        Download a file or directory to a local directory if applicable, and return a
        local path for it.

        Args:
            download_path: (deprecated) Relative source path to the desired files.
            remote_path: Relative source path to the desired files.
            path: Absolute path of the local filesystem destination directory to which to
                download the specified files. This directory must already exist.
                If unspecified, the files will either be downloaded to a new
                uniquely-named directory.
            overwrite: if to overwrite the files at/inside `dst_path` if they exist
            progress: value to show progress bar, defaults to None.

        Returns:
            str: Absolute path of the local filesystem location containing the desired files or folder.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            data_directory = client.get_data_directory_by_fqn(fqn="<your-data_directory-fqn>")
            data_directory.download(download_path="<your-desired-download-path>")
            ```
        """
        self._ensure_not_deleted()
        artifacts_repo = self._get_artifacts_repo()
        if download_path != "" and remote_path != "":
            raise ValueError(
                "Only one of 'download_path' or 'remote_path' should be specified."
            )

        if download_path != "":
            warn(
                "`download_path` is deprecated, please use `remote_path` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            remote_path = download_path

        return artifacts_repo.download_artifacts(
            artifact_path=remote_path,
            dst_path=path,
            overwrite=overwrite,
            progress=progress,
        )

    def update(self):
        """
        Updates the current instance of the DataDirectory.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            data_directory = client.get_data_directory_by_fqn(fqn="<your-data-directory-fqn>")
            data_directory.description = 'This is the new description'
            data_directory.update()
            ```
        """
        self._ensure_not_deleted()
        _dataset = self._mlfoundry_artifacts_api.update_dataset_post(
            update_dataset_request_dto=UpdateDatasetRequestDto(
                fqn=self._dataset.fqn,
                description=self.description,
                dataset_metadata=self.metadata,
            )
        )
        self._dataset = _dataset.dataset
        self._set_mutable_attrs()

    def _ensure_not_deleted(self):
        if self._deleted:
            raise MlFoundryException(
                "Data Directory was deleted, cannot access a deleted version"
            )

    def delete(self, delete_contents: bool = False) -> bool:
        """
        Deletes the current instance of the DataDirectory.

        Args:
            delete_contents: set it to true to delete the contents in storage integration. Default is False.

        Returns:
            True if artifact was deleted successfully

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            dataset = client.get_data_directory_by_fqn(fqn="<your-data-directory-fqn>")
            dataset.delete()
            ```
        """
        self._ensure_not_deleted()
        self._mlfoundry_artifacts_api.delete_dataset_post(
            delete_dataset_request_dto=DeleteDatasetRequestDto(
                id=self._dataset.id, delete_contents=delete_contents
            )
        )
        self._deleted = True
        return True



import functools
import sys
import zipfile
from typing import Dict, Optional

import questionary
import rich_click as click
from packaging.version import parse as parse_version
from requests.exceptions import ConnectionError, Timeout
from rich.markup import escape
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table

from truefoundry.cli.config import CliConfig
from truefoundry.cli.console import console
from truefoundry.common.exceptions import HttpRequestException

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata


_CLICK_VERSION = None


def setup_rich_click():
    click.rich_click.STYLE_ERRORS_SUGGESTION = "blue italic"
    click.rich_click.SHOW_ARGUMENTS = True
    click.rich_click.USE_RICH_MARKUP = True
    click.rich_click.STYLE_HELPTEXT = ""


def handle_exception(exception):
    if CliConfig.debug:
        console.print_exception(show_locals=True)
    if isinstance(exception, HttpRequestException):
        print_dict_as_table_panel(
            {
                "Status Code": str(exception.status_code),
                "Error": escape(exception.message),
            },
            title="Command Failed",
            border_style="red",
        )
    elif isinstance(exception, Timeout):
        loc = ""
        if exception.request:
            loc = f" at {exception.request.url}"
        print_dict_as_table_panel(
            {"Error": f"Request to TrueFoundry{loc} timed out."},
            title="Command Failed",
            border_style="red",
        )
    elif isinstance(exception, ConnectionError):
        loc = ""
        if exception.request:
            loc = f" at {exception.request.url}"
        print_dict_as_table_panel(
            {
                "Error": f"Couldn't connect to TrueFoundry{loc}. Please make sure that the provided `--host` is correct."
            },
            title="Command Failed",
            border_style="red",
        )
    elif isinstance(exception, KeyError):
        # KeyError messages in Python just suck - they tell you the key that was missing, but not the context
        console.print_exception(show_locals=False, max_frames=1)
    else:
        print_dict_as_table_panel(
            {"Error": escape(str(exception))},
            title="Command Failed",
            border_style="red",
        )


def handle_exception_wrapper(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            handle_exception(e)
            sys.exit(1)

    return inner


def print_error(message):
    text = Padding(message, (0, 1))
    console.print(
        Panel(
            text,
            border_style="red",
            title="Command failed",
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def print_message(message):
    text = Padding(message, (0, 1))
    console.print(
        Panel(
            text,
            border_style="cyan",
            title="Success",
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def print_dict_as_table_panel(
    dct: Dict[str, str],
    title: str,
    border_style: str = "cyan",
    key_color: str = "cyan",
):
    table = Table(show_header=False, box=None)
    table.add_column("Key", style=f"bold {key_color}", width=15)
    table.add_column("Value", overflow="fold")
    for key, value in dct.items():
        table.add_row(key, value)
    console.print(
        Panel(
            table,
            border_style=border_style,
            title=title,
            title_align="left",
            width=click.rich_click.MAX_WIDTH,
        )
    )


def unzip_package(path_to_package, destination):
    with zipfile.ZipFile(path_to_package, "r") as zip_ref:
        zip_ref.extractall(destination)


def prompt_if_no_value_and_supported(prompt: str, hide_input: bool = True):
    global _CLICK_VERSION
    kwargs = {}

    if _CLICK_VERSION is None:
        try:
            _CLICK_VERSION = parse_version(importlib_metadata.version("click"))
        except Exception:
            _CLICK_VERSION = parse_version("0.0.0")

    if _CLICK_VERSION.major >= 8:
        kwargs = {"prompt": prompt, "hide_input": hide_input, "prompt_required": False}

    return kwargs


def select_cluster(cluster: Optional[str] = None) -> str:
    """
    Retrieve available clusters and either return the specified one after validation
    or allow the user to interactively select from the list.
    """
    from truefoundry.deploy.lib.clients.servicefoundry_client import (
        ServiceFoundryServiceClient,
    )

    clusters = ServiceFoundryServiceClient().list_clusters()

    if not clusters:
        raise click.ClickException("No clusters found in your account.")

    if cluster:
        if not any(c.id == cluster for c in clusters):
            raise click.ClickException(
                f"Cluster {cluster} not found. Either it does not exist or you might not be authorized to access it"
            )
        return cluster

    choices = {cluster.id: cluster for cluster in clusters}
    selected_cluster = questionary.select(
        "Pick a Cluster:", choices=list(choices.keys())
    ).ask()
    if not selected_cluster:
        raise click.ClickException("No cluster selected.")
    return selected_cluster



from typing import Any, List, Optional


class OutputCallBack:
    def print_header(self, line: Any) -> None:
        print(line)

    def _print_separator(self) -> None:
        print("-" * 80)

    def print_line(self, line: str) -> None:
        print(line)

    def print_lines_in_panel(
        self, lines: List[str], header: Optional[str] = None
    ) -> None:
        self.print_header(header)
        self._print_separator()
        for line in lines:
            self.print_line(line)
        self._print_separator()

    def print_code_in_panel(
        self, lines: List[str], header: Optional[str] = None
    ) -> None:
        self.print_lines_in_panel(lines, header)

    def print(self, line: Any) -> None:
        # just an alias
        self.print_line(line)



TFY = "TFY"

# TODO: probably create another `rich_messages.py` and apply all formatting there
PROMPT_LOGOUT_SUCCESSFUL = """[green bold]Logged Out![/]"""
PROMPT_ALREADY_LOGGED_OUT = """[yellow]You are already logged out[/]"""
PROMPT_DELETED_WORKSPACE = """[green]Deleted workspace {!r}[/]"""
PROMPT_DELETED_APPLICATION = """[green]Deleted Application {!r}[/]"""
PROMPT_NO_WORKSPACES = f"""[yellow]No workspaces found. Either cluster name is wrong, or your cluster doesn't contain any workspaces. You can create one with [bold]{TFY} create workspace[/][/]"""
PROMPT_NO_APPLICATIONS = f"""[yellow]No applications found. You can create one with [bold]{TFY} deploy[/] from within your application folder"""
PROMPT_NO_VERSIONS = """[yellow]No application versions found."""
PROMPT_APPLYING_MANIFEST = """[yellow]Applying manifest for file {!r}[/]"""
PROMPT_DELETING_MANIFEST = """[yellow]Deleting manifest for file {!r}[/]"""



import copy
import difflib
from typing import Any, Dict, Optional

import yaml
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel


def _normalize_manifest_for_diff(obj: Any) -> Any:
    """
    Normalize a manifest object for consistent diffing.
    Only sorts the 'integrations' field by name
    """
    if isinstance(obj, dict):
        _type = obj.get("type") or ""
        integrations = obj.get("integrations")
        if (
            _type.startswith("provider-account/")
            and isinstance(integrations, list)
            and all(isinstance(item, dict) and "name" in item for item in integrations)
        ):
            result = copy.deepcopy(obj)
            result["integrations"] = sorted(integrations, key=lambda x: x["name"])
            return result
    return obj


def format_manifest_for_diff(manifest: Dict[str, Any]) -> str:
    """
    Format a manifest for diffing with consistent formatting.

    Args:
        manifest: The manifest dictionary to format

    Returns:
        A consistently formatted YAML string suitable for diffing
    """
    # Normalize the manifest for diffing
    normalized_manifest = _normalize_manifest_for_diff(manifest)
    return yaml.dump(normalized_manifest, sort_keys=True, indent=2)


def generate_manifest_diff(
    existing_manifest: Dict[str, Any],
    new_manifest: Dict[str, Any],
    manifest_name: str = "manifest",
) -> Optional[str]:
    """
    Generate a unified diff between existing and new manifests.

    Args:
        existing_manifest: The existing manifest of the resource
        new_manifest: The new manifest being applied
        manifest_name: Name of the manifest for diff headers

    Returns:
        Unified diff string if there are differences, None if no differences
    """
    # Format both manifests consistently
    existing_formatted = format_manifest_for_diff(existing_manifest)
    new_formatted = format_manifest_for_diff(new_manifest)

    # Generate diff
    existing_lines = existing_formatted.splitlines(keepends=True)
    new_lines = new_formatted.splitlines(keepends=True)
    diff_lines = list(
        difflib.unified_diff(
            existing_lines,
            new_lines,
            fromfile=f"existing/{manifest_name}",
            tofile=f"new/{manifest_name}",
        )
    )

    if not diff_lines:
        return None

    return "".join(diff_lines)


def print_manifest_diff(
    existing_manifest: Dict[str, Any],
    new_manifest: Dict[str, Any],
    manifest_name: str = "manifest",
    console: Optional[Console] = None,
) -> bool:
    """
    Generate and print a colored diff between manifests.

    Args:
        existing_manifest: The existing manifest of the resource
        new_manifest: The new manifest being applied
        manifest_name: Name of the manifest for diff headers
        console: Optional Rich console instance to use for printing

    Returns:
        True if diff was printed, False if no diff
    """
    if console is None:
        console = Console()

    diff_text = generate_manifest_diff(existing_manifest, new_manifest, manifest_name)

    if diff_text is None:
        console.print(f"[green]No changes detected for {manifest_name}[/]")
        return False

    console.print(
        Panel(Markdown(f"```diff\n{diff_text}\n```"), title=f"Diff for {manifest_name}")
    )
    console.print()
    return True
